"""Helper for adding a projectum to the projecta collection.

Projecta are small bite-sized project quanta that typically will result in
one manuscript.
"""

import datetime as dt

import dateutil.parser as date_parser
from dateutil.relativedelta import relativedelta
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, get_pi_id, get_uuid

# from regolith.schemas import MILESTONE_TYPES


TARGET_COLL = "projecta"


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument("name", help="A short but unique name for the projectum", default=None)
    subpi.add_argument("lead", help="id of the group lead or tbd", default=None)
    subpi.add_argument("-n", "--notes", nargs="+", help="Anything to note", default=[])
    subpi.add_argument("-d", "--description", help="Slightly longer description of the projectum")
    subpi.add_argument(
        "-c",
        "--collaborators",
        nargs="+",
        help="list of outside collaborator ids separated by spaces, "
        "'aeinstein efermi'.  Builders will get the full names "
        "from the contacts collection",
    )
    subpi.add_argument(
        "-m",
        "--group-members",
        nargs="+",
        help="list of group member ids, e.g., 'astudent acolleague'. "
        "Builders will get full names from people collection."
        "Do not add the lead or the group"
        "the pi who are added by default.",
    )
    subpi.add_argument(
        "-g", "--grants", nargs="+", help="grant or (occasionally) list of grants that support this work"
    )
    subpi.add_argument("-u", "--due-date", help="proposed due date for the deliverable", **date_kwargs)
    subpi.add_argument(
        "--checklist", action="store_true", help="Use this to turn the prum into a paper submission" "checklist."
    )
    # Do not delete --database arg
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    # Do not delete --date arg
    subpi.add_argument(
        "--date", help="The begin_date for the projectum  Defaults to " "today's date.", **date_kwargs
    )
    return subpi


class ProjectumAdderHelper(DbHelperBase):
    """Helper for adding a projectum to the projecta collection.

    Projecta are small bite-sized project quanta that typically will result in
    one manuscript.
    """

    # btype must be the same as helper target in helper.py
    btype = "a_projectum"
    needed_colls = [f"{TARGET_COLL}", "groups", "people"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        if not rc.date:
            now = dt.date.today()
        else:
            now = date_parser.parse(rc.date).date()
        if not rc.due_date:
            due_date = now + relativedelta(years=1)
        else:
            due_date = date_parser.parse(rc.due_date).date()
        key = f"{rc.lead[:2]}_{''.join(rc.name.casefold().split()).strip()}"

        coll = self.gtx[rc.coll]
        pdocl = list(filter(lambda doc: doc["_id"] == key, coll))
        if len(pdocl) > 0:
            raise RuntimeError("This entry appears to already exist in the collection")
        else:
            pdoc = {}

        pdoc.update(
            {
                "begin_date": now,
                "log_url": "",
                "supplementary_info_urls": [],
                "name": rc.name,
                "pi_id": rc.pi_id,
                "lead": rc.lead,
                "notes": rc.notes,
            }
        )
        if rc.lead == "tbd":
            pdoc.update({"status": "proposed"})
        else:
            pdoc.update({"status": "started"})

        if rc.description:
            pdoc.update(
                {
                    "description": rc.description,
                }
            )
        if rc.grants:
            if isinstance(rc.grants, str):
                rc.grants = [rc.grants]
            pdoc.update({"grants": rc.grants})
        else:
            pdoc.update({"grants": ["tbd"]})
        if rc.group_members:
            if isinstance(rc.group_members, str):
                rc.group_members = [rc.group_members]
            pdoc.update({"group_members": rc.group_members})
        else:
            pdoc.update({"group_members": []})
        if rc.collaborators:
            if isinstance(rc.collaborators, str):
                rc.collaborators = [rc.collaborators]
            pdoc.update(
                {
                    "collaborators": rc.collaborators,
                }
            )
        else:
            pdoc.update(
                {
                    "collaborators": [],
                }
            )

        pdoc.update({"_id": key})
        pdoc.update(
            {
                "deliverable": {
                    "due_date": due_date,
                    "audience": [
                        "Who will use the software or read the paper? "
                        "Your target audience. e.g., beginning grad in chemistry"
                    ],
                    "success_def": "audience is happy",
                    "scope": [
                        "If this is a software release, list any use-cases that are in "
                        "scope. These may be located in the Gdoc associated with the "
                        "prum.  Otherwise include some other kind of scope description "
                        "of what is in and what is not",
                        "If this is a science paper summarize: ",
                        "the scientific question that is being answered",
                        "any hypotheses that will be tested to answer it",
                        "the approach that will be taken",
                    ],
                    "platform": "description of how and where the audience will access "
                    "the deliverable.  The journal where it will be submitted "
                    "if it is a paper. Whether it is a web-app, or which "
                    "operating systems will be supported, for software",
                    "roll_out": [
                        "steps that the audience will take to access and interact with the deliverable",
                        "leave as empty list for paper submissions",
                    ],
                    "status": "proposed",
                }
            }
        )
        pdoc.update(
            {
                "kickoff": {
                    "due_date": now + relativedelta(days=7),
                    "audience": ["lead", "pi", "group_members"],
                    "name": "Kick off meeting",
                    "objective": "introduce project to the lead",
                    "status": "converged",
                }
            }
        )
        secondm = {
            "due_date": now + relativedelta(days=21),
            "name": "Kickoff meeting",
            "objective": "Prum Lead understands the project deliverables " "and goals",
            "audience": ["lead", "pi", "group_members"],
            "status": "converged",
            "notes": [
                "() Schedule the meeting",
                "() Have the meeting, take good notes",
                "() Update the prum milestones with plan",
            ],
            "progress": {"text": "", "slides_urls": []},
            "type": "meeting",
            "uuid": get_uuid(),
        }
        pdoc.update({"milestones": [secondm]})

        if rc.checklist:
            pdoc = self.insert_checklists(pdoc, now)

        rc.client.insert_one(rc.database, rc.coll, pdoc)

        print(f"{key} has been added in {TARGET_COLL}")

        return

    def insert_checklists(self, pdoc, now):
        """Create manuscript checklist, one item as one milestone."""
        presubmission_checklist = [
            (
                "Create slide figures",
                "Create Inkscape graphics (Inkscape is preferrable over ppt) for the slides "
                "and place in a ``figures`` directory in the slides directory. "
                "These may then be used either in beamer or ppt. "
                "Iterate with PI to convergence. (to get started with Inkscape download and install it, "
                "then run the program and navigate to Help-->Tutorials.  "
                "The first two ('Basic' and 'Shapes') should probably be enough for someone "
                "to get basic functionality.).",
            ),
            (
                "Create slides",
                "Create a 'slides' folder in the paper repo or a Google slides deck for a series of talk slides. "
                "Iterate the slide skeleton with PI to convergence. "
                "(For a beamer template: https://gitlab.thebillingegroup.com/talks/beamerTalkTemplate).",
            ),
            (
                "Create a highlight slide",
                "Create a 'highlight' folder in the paper repo. Create a single 'highlight' slide that describes "
                "the result following NSF/DOE guidelines. Place it in the 'highlight' folder. "
                "Iterate with PI to convergence (highlight templates and examples can be found in "
                "https://gitlab.thebillingegroup.com/papers/highlights)",
            ),
            (
                "Create public-summary",
                "Create a public summary in a text file. Place it in the 'highlight' folder. "
                "Iterate with PI to convergence. (The kudos template and example can be found at "
                "https://docs.google.com/document/d/1j4ZsM8zS_nZo03s7T48uwzDbh8xTPksAQM3ZLgJ-g-Y).",
            ),
            (
                "Add links to urls with supplementary info in the prum",
                "Add url links to raw or analyzed data and code that was used in the data analysis, "
                "and anything else, in gitlab or elsewhere into the 'supplementary_info_urls' field in the prum",
            ),
        ]

        submission_checklist = [
            (
                "Add paper to citations collection in regolith database",
                "If it will be submitted to a preprint server such as arxiv, put it in the citations collection "
                "in a public database, if the paper is not on a preprint server "
                "put the entry in the citations collection on a non-public "
                "(e.g., group) database Check, double check, and triple check "
                "that the tag for the grant is correct and the tags for the facilities are correct. "
                "Any questions, ask PI.",
            ),
            (
                "Create public facing supplementary information as needed",
                "Ask first before making things public, but any code or data "
                "that we want to publish as supplementary info should be put in a place "
                "where it can be made public or at least available to referees "
                "(e.g., in a repo on GitHub).",
            ),
            (
                "Make sure the public facing SI info has the right license",
                "check with the PI for the best license to put on this, "
                "but it will likely be some kind of creative commons license",
            ),
            (
                "Put url links to all open shared SI data repositorities into the 'supplementary_info_urls' field "
                "in the entry for this paper in the citations collection.",
                "the url links in supplementary_info_urls in the projecta collection are to internal data- "
                "and code- stores so future group members can find things. "
                "The fields in the citations collection should be public facing urls",
            ),
            (
                "Check the author list",
                "Check the author list. Last chance to check for missing authors. "
                "Does each author agree to submit to the specific journal? "
                "Make sure all authors have approved submission.",
            ),
            (
                "Check publisher accounts",
                "Check that all of the authors have accounts for the publisher you are submitting to "
                "(if possible) and that they have their ORCID IDs associated with their accounts "
                "(ie. for ACS, authors need to link their Paragon accounts with their ORCID IDs).",
            ),
            (
                "Check institution",
                "Is author's name, institution correct? Last chance to avoid embarrassing typos.",
            ),
            (
                "Check acknowledgement",
                "Are beamlines, grants, fundings properly acknowledged at the end of the paper? "
                "Double check this with PI and use the ackno statements in the Group Google group.",
            ),
            (
                "Check figures and tables",
                "Are all the figures, tables in the paper correct (the ones you intended)?",
            ),
            (
                "Check figure captions",
                "Check the Figure captions for errors. If they refer to a green line, is the relevant line green, "
                "and so on.",
            ),
            (
                "Check figure axis labels",
                "Check that the figure axis labels are correctly labeled. "
                "Make sure it doesn't say G when F is plotted. Make sure the units are correct. "
                "Make sure it says 'G (A^-2)' and NOT 'G(r)' (common mistake).",
            ),
            (
                "Check table captions",
                "Check the table caption is correct. Are all the items in the table "
                "properly defined in the caption. If it is a crystal structure, are the space group "
                "and special positions mentioned in the caption? Is all the info correct?",
            ),
            ("Check numbers in the table", "Check all the numbers in the tables for errors."),
            (
                "Check any question marks",
                "Check all the question marks in the text. Is there any 'FIG.???' or unrecognized character?",
            ),
            (
                "Check figure references",
                "Check references to the all figures and tables. "
                "Does reference to Figure 4 refer to the right figure for example.",
            ),
            (
                "Check references",
                "Go through the references and find all the errors. Correct errors in the bibliographic database "
                "(citations.yml, or the Zotero collection, for example), not in the local bib file. "
                "Did all the journal names compile correctly? Are they all consistently in abbreviated form "
                "(or full form if that is the style, though that is rare). Volume, "
                "year and page numbers appear for all references? Hard to find errors in these numbers, "
                "but when you do, definitely correct the database!",
            ),
            ("Check reference style", "Is reference's style in accordance with journal's requirement?"),
            (
                "Check journal submission requirements",
                "Check the journal submission requirements for cover letters, table of contents pictures, "
                "list of referees, etc..",
            ),
            ("Check arxiv", "Check with PI; will the paper be submitted to arXiv?"),
            (
                "Get approval",
                "Get final approval from all authors for the final version of the manuscript, cover letter, "
                "referees list, etc., submission to arXiv if appropriate.",
            ),
            (
                "Check open access",
                "Check with coauthors if the paper should be submitted open access. "
                "Do not guess on this one as there may be financial consequences",
            ),
            (
                "Check cover letter",
                "In the cover letter, does it contain editor-in-chief's name and institution "
                "(usually at the top left of the letter) ? Is the content of letter concise and eye-catching? "
                "Are (three) suggested reviewers' information in the letter?",
            ),
            (
                "Do a trial submission",
                "Go through the online submission process finding out all the questions that you will be asked, "
                "but don't actually press submit, and then check with the PI about the right answers",
            ),
            ("Commit and push", "Commit all the changes to your local repo, then push the changes to gitlab."),
            (
                "Create a PR of the updated citations entry",
                "Create a PR to merge to the billingeGroup repository and work to get it merged",
            ),
            ("Submit to journal", "Go ahead and make the submission, usually online."),
            (
                "Push a tag",
                "If during the submission process you need to make any changes, do it in your local repo and "
                "make another commit and push. When the submission is finalized, tag the repo that points to "
                "THIS VERSION IS THE SUBMITTED VERSION. create the submission tag.  If the current version of "
                "your local repo is the submitted version, type, e.g., `git tag -l` to list previous tags "
                "(try and keep the tag name formatting consistent) "
                "`git tag -a 20180525PRLsubmitted -m <initial submission to PRL>` `git push origin <tag_name>`.",
            ),
            (
                "Modify tag if needed",
                "If you forgot to tag and made some changes to the repo "
                "and need to point the tag to an earlier version, or want to view all the different tags, "
                "or do some other complicated thing, more info about tagging git repos is here: "
                "https://git-scm.com/book/en/v2/Git-Basics-Tagging",
            ),
            ("Submit to arxiv if needed", "Submit to arxiv if appropriate."),
            (
                "Push an arxiv tag",
                "Make a new tag of the version submitted to arXiv with the name arXivSubmitted20170610.",
            ),
            (
                "Check db errors",
                "In your rg-db-public/local directory, run `regolith build publist --people lyang` "
                "(replace `lyang` with your own name ID in the group) to make sure "
                "that you publist is building properly. Make sure that the publication appears correctly "
                "with no errors and fix anything. If there are problems with the latex building, "
                "run the commands with --no-pdf, which yields the latex source but doesn't build it, then build "
                "the latex manually. The complied tex and pdf files are located in the `_build` folder. "
                "If any problem about installing regolith and databases, please refer to "
                "[rg-db-group wiki]("
                "https://github.com/Billingegroup/rg-db-group/wiki/Set-up-regolith-and-databases).",
            ),
            ("Get arxiv reference", "Wait a day to get the full arXiv reference."),
            ("Email coauthors", "Send an email to coauthors letting them know the arXiv citation information."),
            ("Ask PI if anything unfinished", "Ask PI about any items unfinished."),
            ("Email PI", "Email PI if finish the above."),
        ]

        resubmission_checklist = [
            (
                "Work on the changes",
                "Make changes to the manuscript in the repo. You don't need to save it to a new name or anything "
                "because we can recover the old version by checking out the tagged version.",
            ),
            (
                "Write rebuttal letter",
                "Write a proper rebuttal letter based on reviewers' comments and place in the repo. "
                "In this letter address all of the referee's points, one by one. Place a copy of "
                "the referee's comments in the repo. Give it a unique filename in case "
                "there are more referee comments from later submissions!",
            ),
            (
                "Check author list",
                "This is a good time to check that the author list is correct "
                "(don't need to add anyone or remove anyone) and that the acknowledgements "
                "have been done correctly, the figures are correct, the figure captions and table captions "
                "are correct and all the figures and tables are correctly referenced, and there are "
                "no compilation errors. Check all references for errors and update any "
                "'unpublished' references if they have been published.",
            ),
            (
                "Diff the changes",
                "create a diff.pdf file that shows changes to the manuscript between the version in the tag of "
                "the previous submission and the current version, and include this in the resubmission.",
            ),
            (
                "Send to coauthors",
                "Send the final version to your coauthors. Tell them you will 'submit on' "
                "where is somewhere around 48 hours later and ask for any final corrections etc. from them. "
                "Offer them the chance to extend the deadline if they need more time, "
                "i.e., write 'if you need more time, lease let me know.' "
                "However, it is assumed that all the authors have been involved in the correction process up to "
                "this point so they only have to give it one final thought...",
            ),
            ("Git commit changes", "Commit all the changes to your local repo, then push the changes to gitlab."),
            ("Resubmit", "Resubmit following the instructions of the journal."),
            (
                "Commit any additional changes",
                "If during the submission process you need to make any changes, do it in your local repo and "
                "make another commit and push.",
            ),
            ("Push a resubmission tag", "Make a new resubmission tag (see above for details)"),
            (
                "Check db entry",
                "Check the entry in citations.yml doesn't need to be updated, and update if it does.",
            ),
            ("Ask PI if anything unfinished", "Ask PI about any items unfinished."),
            ("Email PI", "Email PI if finish the above."),
        ]

        accepted_checklist = [
            (
                "Share the news",
                "Congratulations on the acceptance of the paper. Let all coauthors know the great news!",
            ),
            (
                "Share with BNL if needed",
                "If it is a BNL paper (check with PI, but it should acknowledge BNL funding-not just beamtime), "
                "send a pdf copy of the accepted version of the paper from the repo to Arlene at BNL to get "
                "a BNL publication number.  If you are not sure what this means, ask PI",
            ),
            (
                "Share the proof",
                "When you receive the proofs, share them quickly with all the authors. Request comments back "
                "in 24 hours. Proofs should be responded to within 48 hours in normal circumstances.",
            ),
            ("Respond the editor", "Go through and answer any questions from the editor."),
            (
                "Check author names",
                "Last chance to check that all the authors' names are correct and there are no missing authors.",
            ),
            ("Check institutions", "Check all authors' institutions are correct."),
            (
                "Check acknowledgement",
                "Make sure that all funding and all beamlines used are correctly acknowledged. "
                "Usually this is done by the bosses, but more eyes catch more mistakes.",
            ),
            (
                "Update the db entry",
                "In citations.yml, (the reference should have been added during the submission step) double check "
                "the grants{}, facilities{}, nb{} field entries. Any questions, ask PI. Put 'to be published' "
                "in the note{} section. If it has not been submitted to arxiv before, move the entry "
                "from rg-db-group to rg-db-public github repo. Otherwise, it should be at rg-db-public already. "
                "Create a PR to merge to the billingeGroup repository for edits if necessary.",
            ),
            (
                "Check db errors",
                "In your rg-db-public/local directory, run `regolith build publist --people lyang` "
                "(replace lyang with your name) to make sure that you publist is building properly. "
                "Make sure that the publication appears correctly with no errors and fix anything. If there are "
                "problems with the latex building, run the commands with --no-pdf, "
                "which yields the latex source but doesn't build it, then build the latex manually. "
                "If any problem about installing regolith and databases, please refer to "
                "[rg-db-group wiki]("
                "https://github.com/Billingegroup/rg-db-group/wiki/Set-up-regolith-and-databases).",
            ),
            (
                "Check figures and tables",
                "Are all the figures, tables in the paper correct (the ones you intended)?",
            ),
            (
                "Check the figure caption",
                "Check the Figure captions for errors. If they refer to a green line, is the relevant line green, "
                "and so on.",
            ),
            (
                "Check figure axis labels",
                "Check that the figure axis labels are correctly labeled. Make sure it doesn't say G when "
                "F is plotted. Make sure the units are correct. Make sure it says 'G (A^-2)' and NOT 'G(r)' "
                "(common mistake)",
            ),
            (
                "Check table captions",
                "Check the table caption is correct. Are all the items in the table "
                "properly defined in the caption. If it is a crystal structure, are the space group "
                "and special positions mentioned in the caption? Is all the info correct?",
            ),
            ("Check numbers in the table", "Check all the numbers in the tables for errors."),
            (
                "Check figure references",
                "Check references to the all figures and tables. Does reference to Figure 4 refer to "
                "the right figure for example",
            ),
            (
                "Check references",
                "Go through the references and find all the errors. Correct errors in the bibliographic database "
                "AS WELL AS on the proofs the manuscript. Did all the journal names compile correctly? "
                "Are they all consistently in abbreviated form "
                "(or full form if that is the style, though that is rare). Volume, year and page numbers "
                "appear for all references? Hard to find errors in these numbers, but when you do, "
                "definitely correct the database!",
            ),
            (
                "Check unpublished references",
                "If any references are listed as unpublished, on arXiv or submitted or something, "
                "check if they have appeared and give the full reference if at all possible. To do this, "
                "update the bibliographic database (e.g., citations.yml) with this information and then "
                "recompile the references, then copy paste the new bbl file back into the TeX source.",
            ),
            (
                "Check reference titles if needed",
                "If the manuscript style has titles in the references, make sure there are no capitalization "
                "or other compilation errors. Again, correct these in the database using {braces} around words "
                "where you want to preserve the capitalization as well as on the proof.",
            ),
            (
                "Read the paper",
                "Finally, after you have done all these 'mechanical' checks, "
                "read through the paper and try and find any typos or other problems. "
                "Resist the temptation to do any rewriting here...you are looking for mispellings "
                "and missing or extra words and so on.",
            ),
            (
                "Apply corrections from coauthors",
                "Collect all the corrections from the other authors "
                "and add any additional ones to the proof and return it.",
            ),
            ("Email coauthors", "Send an email to your coauthors that this was successfully resubmitted."),
            (
                "Revisit talk slides",
                "Revisit the set of talk slides that summarize the result in a few slides "
                "if they need to be updated. Iterate with PI to convergence.",
            ),
            (
                "Revisit the highlight slide",
                "Create a single 'highlight' slide that describes the result following NSF/DOE guidelines. "
                "Place it in the 'highlight' folder. Iterate with PI to convergence "
                "(highlight templates and examples can be found in "
                "https://gitlab.thebillingegroup.com/highlights/highlightTemplate)",
            ),
            (
                "Check the url links to all open shared SI data repositorities "
                "in the 'supplementary_info_urls' field in the entry for this paper "
                "in the citations collection. make sure they work and the data are accessible.",
            ),
            (
                "Create web news",
                "Create a web news story for thebillingegroup.com site. Place it in the 'highlight' folder. "
                "Iterate with PI to convergence",
            ),
            (
                "Revisit kudos",
                "Revisit the Kudos summary if it needs to be updated. Iterate with PI to convergence.",
            ),
            ("Ask PI if anything unfinished", "Ask PI about any items unfinished."),
            ("Email PI", "Email PI if finish the above."),
        ]

        published_checklist = [
            ("Congrats", "Phew, it is over! Pat yourself on the back and celebrate!"),
            ("Let coauthors know", "Let your coauthors know the link to the final paper and the final reference."),
            (
                "Update db entry",
                "Update citations.yml at rg-db-public github repo with the correct reference information. "
                "Commit your edited citations.yml and create a PR to merge to the billingeGroup repository.",
            ),
            (
                "Check db entry",
                "CAREFULLY double and triple check the meta-data associated with the paper in citations.yml:",
            ),
            (
                "Check grants in the db entry",
                "grant{} lists just the billinge-group grants that appeared in the acknowledgement section. "
                "They have standard abbreviations that are listed at the top of the citations.yml file, "
                "e.g., fwp, EFRC10, etc.. Use the right standard or the whole system becomes broken! "
                "If not sure.....ask PI. List all grants in a comma-separated list.",
            ),
            (
                "Check the facility in the db entry",
                "facility{} is every beamline that was used for data collection. "
                "Again, use the standard abbreviations at the top of the file. "
                "Use two levels of granularity for each, so X17A would be: 'nsls, x17a', "
                "if X17A and X7B were used it would be 'nsls, x17a, x7b' and so on.",
            ),
            (
                "Check the nb in the db entry nb is some other tags, "
                "also listed at the top of the file. 'art' for a regular article and 'hilite' "
                "if it is one of our top top papers are the most common.",
            ),
            (
                "Check the tags in the db entry",
                "tags should reflect the content so we can automatically build reading lists by subject. "
                "Most papers are PDF papers, so no need to say pdf, be more targeted.",
            ),
            (
                "Add the url link to the SI at the journal to the 'supplementary_info_urls' field "
                "in the entry for this paper in the citations collection.",
                "Don't necessarily remove the others if they are still open and public. "
                "It is also ok to put URLs to our data repositories as part of the SI in the paper",
            ),
            (
                "Check db errors",
                "In your rg-db-public/local directory, run `regolith build publist --people lyang` "
                "(replace lyang with your name) to make sure that you publist is building properly. "
                "Make sure that the publication appears correctly with no errors and fix anything. If there are "
                "problems with the latex building, run the commands with --no-pdf, which yields the latex source "
                "but doesn't build it, then build the latex manually.",
            ),
            (
                "Add/update to Zotero",
                "Add or update the published reference to the billinge-group-bib folder "
                "in our group Zotero account",
            ),
            (
                "Finalize the highlight slide",
                "Make a highlight of your work and put it in gitlab/highlights "
                "(if not done already during the accepted paper checklist). "
                "Look in there for standards to work from. This is an important activity. "
                "Now you have done your great work, this is how you can advertise it to others. "
                "Top papers we send these highlights to the funding agencies. "
                "Iterate the highlight with PI till it is converged.",
            ),
            (
                "Finalize figures and talk slides",
                "Make figures and talk slides that will be used in talks "
                "and place these on gitlab on talks/figures. Iterate this with PI till it is converged.",
            ),
            (
                "Update arXiv if necessary",
                "If the paper was listed on a preprint server like arXiv, submit a note to arXiv that the paper "
                "has appeared and give the full reference. If the journal copyright allows "
                "you can post the published version here, but normally that is not alllowed! "
                "Still, it is important that people who find the paper on arXiv "
                "get directed to the correct reference.",
            ),
            ("Ask PI if anything unfinished", "Ask PI about any items unfinished."),
            ("Email PI", "Email PI if finish the above."),
        ]

        checklistm_list = []
        checklist_delay_days = (
            [7] * len(presubmission_checklist)
            + [14] * len(submission_checklist)
            + [74] * len(resubmission_checklist)
            + [134] * len(accepted_checklist)
            + [194] * len(published_checklist)
        )
        checklist_names = (
            ["presubmission"] * len(presubmission_checklist)
            + ["submission"] * len(submission_checklist)
            + ["resubmission"] * len(resubmission_checklist)
            + ["accepted"] * len(accepted_checklist)
            + ["published"] * len(published_checklist)
        )
        checklists = presubmission_checklist + submission_checklist + accepted_checklist + published_checklist
        for (name, objective), checklist_name, delay_days in zip(
            checklists, checklist_names, checklist_delay_days
        ):
            checklistm = {
                "due_date": now + relativedelta(days=delay_days),
                "name": f"{checklist_name} - {name}",
                "objective": objective,
                "audience": [],
                "notes": [],
                "status": "converged",
                "type": "mergedpr",
            }
            checklistm_list.append(checklistm)

        pdoc.update({"milestones": checklistm_list})

        # update the deliverable to fit checklist prum
        pdoc.update(
            {
                "deliverable": {
                    "due_date": now + relativedelta(days=checklist_delay_days[-1]),
                    "audience": ["PI"],
                    "success_def": "audience is happy",
                    "scope": ["checklist", "All publication data and metadata are correct and complete"],
                    "platform": "regolith publication collection in rg-db-public",
                    "roll_out": ["PI merging PRs"],
                    "status": "converged",
                }
            }
        )

        # update the kickoff to fit checklist prum
        pdoc.update(
            {
                "kickoff": {
                    "due_date": now,
                    "audience": ["lead", "pi", "group_members"],
                    "name": "Kick off meeting",
                    "objective": "introduce project to the lead",
                    "status": "finished",
                }
            }
        )
        return pdoc

        # secondm = {'due_date': now + relativedelta(days=21),
        #            'name': 'Example milestone',
        #            'objective': 'This acts as an example milestone. It shows how '
        #                         'to construct your milestones.  It should be deleted '
        #                         'before depositing the prum.',
        #            'audience': ["update the audience as required, but it is "
        #                         "often lead, pi and group_members (see below). "
        #                         "these resolve to their values defined elsewhere.  "
        #                         "If putting in people's names use their ids where "
        #                         "possible and they will be looked for in the "
        #                         "people and contacts collections.",
        #                         'lead', 'pi', 'group_members'],
        #            'status': 'proposed',
        #            'notes': ["() get a clear picture of the deliverable and what "
        #                      "needs to be done to get there.  Have discussions "
        #                      "with group members to develop ideas.  Propose them "
        #                      "even if you are not sure, you will get feedback. "
        #                      "This is the process.",
        #                      "() edit the template 'deliverable' to reflect that.",
        #                      "() chart out milestones at a high level first working "
        #                      "back from the deliverable.  'To deliver that "
        #                      "(deliverable) I have to do this, this and this, and "
        #                      "to deliver it by then I have to do these things by "
        #                      "these dates.'",
        #                      f"() then write the detailed milestones, turning each "
        #                      f"one into a deliverable that is captured in the type "
        #                      f"field.  Allowed values for milestone types are: "
        #                      f"{*MILESTONE_TYPES,}.  If you are tempted to use 'other' "
        #                      f"think more about the milestone objective, you "
        #                      f"probably haven't defined it well enough in your head.",
        #                      "() Think carefully about the deliverable date.  "
        #                      "Can you hit it?  If not, change it so you can hit "
        #                      "it.  You may have to adjust your final deliverable date "
        #                      "but discuss this with the team, because scope could also "
        #                      "be reduced rather than pushing off final deliverable.  "
        #                      "When you are comfortable with the due date, and it has "
        #                      "been reviewed, make the appoint on the calendar, for "
        #                      "example, if it is a meeting or presentation, send "
        #                      "schedule it now.",
        #                      "() within each milestone, make a todo list (this "
        #                      "is a todo-list) in the notes field.  A todo is a "
        #                      "note prepended by () (no space in the middle).  These are the smaller "
        #                      "tasks you will have to do to complete the milestone.  "
        #                      "When you complete them turn () in to (x).",
        #                      "() Iterate all these with your team/advisor maybe "
        #                      "in a Google doc or on Slack to make sure you are "
        #                      "on the right track.  To converge this shoot for "
        #                      "multiple round-trips per day.",
        #                      "() using u_milestones in helper_gui, or "
        #                      "otherwise, get them entered into the prum and the "
        #                      "prum deposited.",
        #                      "() by the time the edited prum is submitted, set "
        #                      "all the statuses to converged or started, and "
        #                      "delete this example milestone.",
        #                      "() later, when you finish a milestone use "
        #                      "u_milestones or editing in the yaml "
        #                      "file set status to finished and add an end_date, "
        #                      "and fill in the progress section. This should be "
        #                      "brief but just links to a brief text description "
        #                      "of the outcome that would be suitable for a grant "
        #                      "annual report, and any url links to "
        #                      "presentations or related docs.  Link to specific "
        #                      "files of relevance to this milestone not general docs and repos.",
        #            ],
        #            "progress": {'text': 'write text here capturing how the milestone '
        #                                 'is progressing, but at the least when the milestone '
        #                                 'closes.  The goal of this is that it will be '
        #                                 'printed as a progress report for a grant '
        #                                 'so write it as if it will be read by an external '
        #                                 'person. It doesnt have to describe the whole prum '
        #                                 'but clearly show the progress that has been made. '
        #                                 'It can be multiple paragraphs or a short statement'
        #                                 'depending on the situation',
        #                         'slides_urls': ["<replace with a URL to, for example, "
        #                                        "a Gslides slide deck "
        #                                        "with useful figures in it, or a Gdoc "
        #                                        "with more complicated info like tables "
        #                                        "or something. This is for the PI to "
        #                                        "be able to find quality content to augment "
        #                                        "the report>","<replace with another URL if "
        #                                                      "more than one is needed>"]},
        #            'type': 'meeting',
        #            'uuid': get_uuid()
        # }



"""Builder for Reading Lists."""

from habanero import Crossref

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.fsclient import _id_key
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection, get_formatted_crossref_reference


class ReadingListsBuilder(LatexBuilderBase):
    """Build reading lists from database entries"""

    btype = "readinglists"
    needed_colls = ["people", "reading_lists"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        self.cr = Crossref()
        rc.verbose = True
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["reading_lists"] = sorted(all_docs_from_collection(rc.client, "reading_lists"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""

        # build the collection of formatted references so that we only go
        # and fetch the formatted references once per doi
        dois, formatted_refs = [], {}
        for rlist in self.gtx["reading_lists"]:
            for paper in rlist["papers"]:
                dois.append(paper.get("doi", ""))
        dois = list(set(dois))
        for item in ["tbd", ""]:
            if item in dois:
                dois.remove("tbd")
                dois.remove("")
        for doi in dois:
            ref_and_date = get_formatted_crossref_reference(doi)
            formatted_refs.update({doi: ref_and_date})

        # loop through the reading lists to build the files
        for rlist in self.gtx["reading_lists"]:
            listid = rlist["_id"]
            outfile_bib = listid

            n = 1
            for paper in rlist["papers"]:
                doi = paper.get("doi")
                paper["text"] = paper["text"].strip(".").strip()
                if doi == "tbd" or doi == "":
                    doi = None
                url = paper.get("url")
                if doi:
                    paper.update(
                        {
                            "reference": formatted_refs.get(doi)[0],
                            "ref_date": formatted_refs.get(doi)[1],
                            "n": n,
                            "label": "DOI",
                        }
                    )
                    n += 1
                elif url:
                    paper["doi"] = url
                    paper.update({"n": n, "label": "URL"})
                    n += 1
                else:
                    paper.update({"n": n})
                    n += 1

            self.render(
                "rlistbibfile.txt",
                outfile_bib + ".txt",
                rlist=rlist,
            )



import copy
import json
import os
import subprocess
import sys
from pathlib import Path

import pytest

from regolith.database import connect
from regolith.dates import convert_doc_iso_to_date
from regolith.main import main
from regolith.mongoclient import load_mongo_col
from regolith.runcontrol import DEFAULT_RC, load_rcfile
from regolith.tests.conftest import ALTERNATE_REGOLITH_MONGODB_NAME, FS_DB_NAME

BILLINGE_TEST = False  # special tests for Billinge group, switch it to False before push to remote


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
def test_fs_to_mongo(make_db, make_mongodb):
    if make_mongodb is False:
        pytest.skip("Skipping, Mongoclient failed to start")
    if BILLINGE_TEST:
        repo = str(Path(__file__).parent.parent.parent.joinpath("rg-db-group", "local"))
    else:
        repo = make_db
    cp1 = subprocess.run(["regolith", "fs-to-mongo"], cwd=repo)
    assert cp1.returncode == 0


def test_fs_to_mongo_python(make_fs_to_mongo_migration_db, make_mongodb):
    if make_mongodb is False:
        pytest.skip("Mongoclient failed to start")
    if BILLINGE_TEST:
        repo = str(Path(__file__).parent.parent.parent.joinpath("rg-db-group", "local"))
    else:
        repo = make_fs_to_mongo_migration_db
    os.chdir(repo)
    try:
        main(["fs-to-mongo"])
    except Exception as e:
        print(e)
        assert False
    else:
        assert True
    replace_rc_dbs(repo)
    os.chdir(repo)
    rc = copy.copy(DEFAULT_RC)
    rc._update(load_rcfile("regolithrc.json"))
    with connect(rc) as rc.client:
        fs_db = rc.client[FS_DB_NAME]
        mongo_db = rc.client[ALTERNATE_REGOLITH_MONGODB_NAME]
        for coll in fs_db.keys():
            original_fs_collection = fs_db[coll]
            migrated_mongo_collection = load_mongo_col(mongo_db[coll])
            for k, v in migrated_mongo_collection.items():
                migrated_mongo_collection[k] = convert_doc_iso_to_date(v)
            for k, v in original_fs_collection.items():
                original_fs_collection[k] = convert_doc_iso_to_date(v)
            assert original_fs_collection == migrated_mongo_collection


def test_mongo_to_fs_python(make_mongo_to_fs_backup_db, make_mongodb):
    if make_mongodb is False:
        pytest.skip("Mongoclient failed to start")
    repo = make_mongo_to_fs_backup_db
    os.chdir(repo)
    try:
        main(["mongo-to-fs"])
    except Exception as e:
        print(e)
        assert False
    else:
        assert True
    replace_rc_dbs(repo)
    os.chdir(repo)
    rc = copy.copy(DEFAULT_RC)
    rc._update(load_rcfile("regolithrc.json"))
    with connect(rc) as rc.client:
        fs_db = rc.client[FS_DB_NAME]
        mongo_db = rc.client[ALTERNATE_REGOLITH_MONGODB_NAME]
        for coll in mongo_db.list_collection_names():
            migrated_fs_collection = fs_db[coll]
            original_mongo_collection = load_mongo_col(mongo_db[coll])
            assert migrated_fs_collection == original_mongo_collection


def replace_rc_dbs(repo):
    cwd = os.getcwd()
    os.chdir(repo)
    with open("regolithrc.json", "r+") as f:
        data = json.load(f)
        data["databases"] = [
            {
                "name": FS_DB_NAME,
                "url": repo,
                "public": True,
                "path": "db",
                "local": True,
                "backend": "filesystem",
            },
            {
                "name": ALTERNATE_REGOLITH_MONGODB_NAME,
                "url": "localhost",
                "path": repo,
                "public": True,
                "local": True,
                "backend": "mongodb",
            },
        ]
        f.seek(0)
        json.dump(data, f, indent=4)
        f.truncate()
    os.chdir(cwd)



"""Helper for adding a to_do task to todos.yml"""

import datetime as dt

import dateutil.parser as date_parser
from dateutil.relativedelta import relativedelta
from gooey import GooeyParser

from regolith.chained_db import _convert_to_dict
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, fragment_retrieval, get_pi_id, get_uuid

TARGET_COLL = "todos"
ALLOWED_IMPORTANCE = [3, 2, 1, 0]


def subparser(subpi):
    date_kwargs = {}
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"
        int_kwargs["widget"] = "IntegerField"
        int_kwargs["gooey_options"] = {"min": 0, "max": 10000}
    subpi.add_argument(
        "description",
        help="the description of the to_do task. If the description has more than one "
        "word, please enclose it in quotation marks.",
        default=None,
    )
    subpi.add_argument(
        "due_date",
        help="Due date of the task, in days from today (integer), or " "as a date in iso format (yyyy-mm-dd)",
    )
    subpi.add_argument(
        "duration",
        help="The estimated duration the task will take in minutes (integer) "
        "e.g., 60 would be a duration of 1 hour.",
    )
    subpi.add_argument(
        "-d",
        "--deadline",
        action="store_true",
        help="specify if the due date (above) has a hard deadline",
    )
    subpi.add_argument(
        "-m",
        "--importance",
        choices=ALLOWED_IMPORTANCE,
        type=int,
        help="The importance of the task. "
        "Corresponds roughly to (3) tt, (2) tf, (1) ft, (0) ff in the Eisenhower matrix of "
        "importance vs. urgency.  An important and urgent task would be 3.",
        default=1,
    )
    subpi.add_argument(
        "-t",
        "--tags",
        nargs="+",
        help="Tags to be associated with this task.  Enter as single words separated by spaces. "
        "The todo list can be filtered by these tags",
    )
    subpi.add_argument(
        "-i",
        "--milestone_uuid",
        help="Add this todo as a task to the projectum milestone with this uuid"
        "Takes a full or partial milestone uuid.",
    )
    subpi.add_argument(
        "-n",
        "--notes",
        nargs="+",
        help="Additional notes for this task. Each note should be enclosed "
        "in quotation marks and different notes separated by spaces",
    )
    subpi.add_argument(
        "-a",
        "--assigned-to",
        help="ID of the group member to whom the task is assigned. Default is the id saved in user.json. ",
    )
    subpi.add_argument(
        "-b",
        "--assigned-by",
        help="ID of the member that is assigning the task. Default is the id saved in user.json. ",
    )
    subpi.add_argument("--begin-date", help="Begin date of the task. Default is today.", **date_kwargs)
    subpi.add_argument(
        "--database",
        help="The database in which the collection will be updated. "
        "Defaults to the first database in regolithrc.json if not "
        "specified.",
    )
    subpi.add_argument(
        "--date",
        help="Enter a date such that the helper can calculate how many days are left "
        "from that date to the deadline. Default is today.",
        **date_kwargs,
    )

    return subpi


class TodoAdderHelper(DbHelperBase):
    """Helper for adding a todo task to todos.yml"""

    # btype must be the same as helper target in helper.py
    btype = "a_todo"
    needed_colls = [f"{TARGET_COLL}", "projecta"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)

        rc.coll = f"{TARGET_COLL}"
        rc.col2 = "projecta"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["projecta"] = sorted(all_docs_from_collection(rc.client, rc.col2), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        if not rc.assigned_to:
            try:
                rc.assigned_to = rc.default_user_id
            except AttributeError:
                print(
                    "Please set default_user_id in '~/.config/regolith/user.json', "
                    "or you need to enter your group id in the command line"
                )
                return
        person = rc.client.find_one(rc.database, rc.coll, {"_id": rc.assigned_to})
        if not person:
            raise TypeError(f"The id {rc.assigned_to} can't be found in the todos collection")
        if not rc.assigned_by:
            rc.assigned_by = rc.default_user_id
        find_person = rc.client.find_one(rc.database, rc.coll, {"_id": rc.assigned_by})
        if not find_person:
            raise TypeError(f"The id {rc.assigned_by} can't be found in the todos collection")
        now = dt.date.today()
        if not rc.begin_date:
            begin_date = now
        else:
            begin_date = date_parser.parse(rc.begin_date).date()
        if not rc.date:
            today = now
        else:
            today = date_parser.parse(rc.date).date()
        try:
            relative_day = int(rc.due_date)
            due_date = today + relativedelta(days=relative_day)
        except ValueError:
            due_date = date_parser.parse(rc.due_date).date()
        if begin_date > due_date:
            raise ValueError("begin_date can not be after due_date")
        if rc.importance not in ALLOWED_IMPORTANCE:
            raise ValueError(f"importance should be chosen from {ALLOWED_IMPORTANCE}")
        else:
            importance = int(rc.importance)

        todolist = person.get("todos", [])
        if not rc.deadline:
            rc.deadline = False
        todo_uuid = get_uuid()
        todolist.append(
            {
                "description": rc.description,
                "uuid": todo_uuid,
                "due_date": due_date,
                "begin_date": begin_date,
                "deadline": rc.deadline,
                "duration": float(rc.duration),
                "importance": importance,
                "status": "started",
                "assigned_by": rc.assigned_by,
            }
        )
        if rc.notes:
            todolist[-1]["notes"] = rc.notes
        if rc.tags:
            todolist[-1]["tags"] = rc.tags
        if rc.milestone_uuid:
            # get the prum that contains the milestone that has the uuid rc.milestone_uuid
            target_prum = fragment_retrieval(self.gtx["projecta"], ["milestones"], rc.milestone_uuid)
            if not target_prum:
                raise RuntimeError(
                    f"No milestone ids were found that match your entry ({rc.milestone_uuid}).\n"
                    "Make sure you have entered the correct milestone uuid or uuid fragment and rerun the helper."
                )
            # checks if the same id is found in different prums
            if len(target_prum) > 1:
                raise RuntimeError(
                    f"Multiple milestone ids match your entry ({rc.milestone_uuid}).\n"
                    "Try entering more characters of the uuid and rerunning the helper."
                )
            target_prum = _convert_to_dict(target_prum[0])
            # now look for milestones that match
            target_mil = [mil for mil in target_prum["milestones"] if rc.milestone_uuid in mil.get("uuid")]
            # Checks if the same id fragment is found in multiple milestones
            # within the same prm
            if len(target_mil) > 1:
                raise RuntimeError(
                    f"Multiple milestone ids match your entry ({rc.milestone_uuid}).\n"
                    "Try entering more characters of the uuid and rerunning the helper."
                )
            else:
                mil = target_mil[0]
                if not mil.get("tasks"):
                    mil.update({"tasks": []})
                mil["tasks"].append(todo_uuid)
                print(
                    f"The milestone uuid {rc.milestone_uuid} in projectum {target_prum.get('_id')} has been "
                    "updated in projecta."
                )
            rc.client.update_one(rc.database, "projecta", {"_id": target_prum.get("_id")}, target_prum)
        indices = [todo.get("running_index", 0) for todo in todolist]
        todolist[-1]["running_index"] = max(indices) + 1
        rc.client.update_one(rc.database, rc.coll, {"_id": rc.assigned_to}, {"todos": todolist}, upsert=True)
        print(f'The task "{rc.description}" for {rc.assigned_to} has been added in {TARGET_COLL} collection.')

        return



"""Copyright (c) 2017, Anthony Scopatz
All rights reserved."""

import json
import os
import tempfile
from copy import deepcopy
from subprocess import CalledProcessError

import pytest
from pymongo import MongoClient
from pymongo import errors as mongo_errors
from xonsh.lib import subprocess
from xonsh.lib.os import rmtree

from regolith.fsclient import dump_yaml
from regolith.schemas import EXEMPLARS

OUTPUT_FAKE_DB = False  # always turn it to false after you used it
# Currently the first two must be named test solely to match the helper map test output text
REGOLITH_MONGODB_NAME = "test"
FS_DB_NAME = "test"
ALTERNATE_REGOLITH_MONGODB_NAME = "mongo_test"


@pytest.fixture(scope="session")
def make_db():
    """A test fixutre that creates and destroys a git repo in a temporary
    directory.
    This will yield the path to the repo.
    """
    cwd = os.getcwd()
    name = "regolith_fake"
    repo = os.path.join(tempfile.gettempdir(), name)
    if os.path.exists(repo):
        rmtree(repo)
    subprocess.run(["git", "init", repo])
    os.chdir(repo)
    with open("README", "w") as f:
        f.write("testing " + name)
    with open("regolithrc.json", "w") as f:
        json.dump(
            {
                "default_user_id": "sbillinge",
                "groupname": "ERGS",
                "databases": [
                    {
                        "name": "test",
                        "url": repo,
                        "public": True,
                        "path": "db",
                        "local": True,
                        "backend": "filesystem",
                    }
                ],
                "repos": [
                    {
                        "_id": "talk_repo",
                        "params": {"namespace_id": "35", "initialize_with_readme": "true", "name": "repo name"},
                        "url": "https://example.com",
                        "api_route": "/url/example",
                        "namespace_name": "talks",
                    }
                ],
                "stores": [
                    {
                        "name": "store",
                        "url": repo,
                        "path": repo,
                        "public": True,
                    }
                ],
                "tokens": [{"_id": "gitlab_private_token", "token": "<private_token>"}],
            },
            f,
        )
    fspath = os.path.join(repo, "db")
    os.mkdir(fspath)
    exemplars_to_fs(fspath)
    subprocess.run(["git", "add", "."])
    subprocess.run(["git", "commit", "-am", "Initial readme"])
    yield repo
    os.chdir(cwd)
    if not OUTPUT_FAKE_DB:
        rmtree(repo)


@pytest.fixture(scope="module")
def make_mongodb():
    """A test fixture that creates and destroys a git repo in a temporary
    directory, as well as a mongo database.
    This will yield the path to the repo.
    """
    forked = False
    name = "regolith_mongo_fake"
    repo = os.path.join(tempfile.gettempdir(), name)
    if os.path.exists(repo):
        rmtree(repo)
    subprocess.run(["git", "init", repo])
    os.chdir(repo)
    with open("README", "w") as f:
        f.write("testing " + name)
    mongodbpath = os.path.join(repo, "dbs")
    os.mkdir(mongodbpath)
    with open("regolithrc.json", "w") as f:
        json.dump(
            {
                "groupname": "ERGS",
                "databases": [
                    {
                        "name": REGOLITH_MONGODB_NAME,
                        "url": "localhost",
                        "path": repo,
                        "public": True,
                        "local": True,
                        "backend": "mongodb",
                    }
                ],
                "stores": [
                    {
                        "name": "store",
                        "url": repo,
                        "path": repo,
                        "public": True,
                    }
                ],
                "mongodbpath": mongodbpath,
            },
            f,
        )
    if os.name == "nt":
        # If on windows, the mongod command cannot be run with the fork or syslog options.
        # Instead, it is installed as a service and the exceptions that would typically be log outputs
        # are handled by the exception handlers below. In addition, the database must always be manually
        # deleted from the windows mongo instance before running a fresh test.
        cmd = ["mongo", REGOLITH_MONGODB_NAME, "--eval", "db.dropDatabase()"]
        try:
            subprocess.check_call(cmd, cwd=repo)
        except CalledProcessError:
            print(
                "Mongodb likely has not been installed as a service. In order to run mongodb tests, make sure\n"
                "to install the mongodb community edition with the following link: \n"
                "https://docs.mongodb.com/manual/installation/"
            )
            yield False
            return
        cmd = ["mongostat", "--host", "localhost", "-n", "1"]
    else:
        cmd = ["mongod", "--fork", "--syslog", "--dbpath", mongodbpath]
        forked = True
    try:
        subprocess.check_call(cmd, cwd=repo)
    except CalledProcessError:
        print(
            "If using linux or mac, Mongod command failed to execute. "
            "If using windows, the status of mongo could \n not be retrieved. "
            "In order to run mongodb tests, make sure to install the mongodb community edition with "
            "\nthe following link:\n"
            "https://docs.mongodb.com/manual/installation/"
        )
        yield False
        return
    try:
        exemplars_to_mongo(REGOLITH_MONGODB_NAME)
    except ConnectionError:
        yield False
        return
    yield repo
    cmd = ["mongo", REGOLITH_MONGODB_NAME, "--eval", "db.dropDatabase()"]
    try:
        subprocess.check_call(cmd, cwd=repo)
    except CalledProcessError:
        print(
            f'Deleting the test database failed, insert "mongo {REGOLITH_MONGODB_NAME} --eval '
            f'"db.dropDatabase()"" into command line manually'
        )
    shut_down_fork(forked, repo)
    if not OUTPUT_FAKE_DB:
        rmtree(repo)


@pytest.fixture(scope="module")
def make_mixed_db():
    """A test fixture that creates and destroys a git repo in a temporary
    directory, as well as a mongo database.
    This will yield the path to the repo.

    This specific test fixture points to a repo that mixes mongo and filesystem backends for the assignments and
    abstracts test collections in EXEMPLARS respectively.
    """
    cwd = os.getcwd()
    forked = False
    name = "regolith_mongo_fake"
    repo = os.path.join(tempfile.gettempdir(), name)
    if os.path.exists(repo):
        rmtree(repo)
    subprocess.run(["git", "init", repo])
    os.chdir(repo)
    with open("README", "w") as f:
        f.write("testing " + name)
    mongodbpath = os.path.join(repo, "dbs")
    os.mkdir(mongodbpath)
    fspath = os.path.join(repo, "db")
    os.mkdir(fspath)
    with open("regolithrc.json", "w") as f:
        json.dump(
            {
                "groupname": "ERGS",
                "databases": [
                    {
                        "name": REGOLITH_MONGODB_NAME,
                        "url": "localhost",
                        "path": repo,
                        "public": True,
                        "local": True,
                        "backend": "mongodb",
                    },
                    {
                        "name": FS_DB_NAME,
                        "url": repo,
                        "public": True,
                        "path": "db",
                        "local": True,
                        "backend": "filesystem",
                    },
                ],
                "stores": [
                    {
                        "name": "store",
                        "url": repo,
                        "path": repo,
                        "public": True,
                    }
                ],
                "mongodbpath": mongodbpath,
            },
            f,
        )
    if os.name == "nt":
        # If on windows, the mongod command cannot be run with the fork or syslog options.
        # Instead, it is installed as a service and the exceptions that would typically be log outputs
        # are handled by the exception handlers below. In addition, the database must always be manually
        # deleted from the windows mongo instance before running a fresh test.
        cmd = ["mongo", REGOLITH_MONGODB_NAME, "--eval", "db.dropDatabase()"]
        try:
            subprocess.check_call(cmd, cwd=repo)
        except CalledProcessError:
            print(
                "Mongod likely has not been installed as a service. In order to run mongodb tests, make sure\n"
                "to install the mongodb community edition with the following link: \n"
                "https://docs.mongodb.com/manual/installation/"
            )
            yield False
            return
        cmd = ["mongostat", "--host", "localhost", "-n", "1"]
    else:
        cmd = ["mongod", "--fork", "--syslog", "--dbpath", mongodbpath]
        forked = True
    try:
        subprocess.check_call(cmd, cwd=repo)
    except CalledProcessError:
        print(
            "If on linux/mac, Mongod command failed to execute. If on windows, the status of mongo could not be\n"
            "retrieved. In order to run mongodb tests, make sure to install the mongodb community edition with\n"
            "the following link:\n"
            "https://docs.mongodb.com/manual/installation/"
        )
        yield False
        return
    # Write one collection doc in mongo
    mongo_coll = "assignments"
    try:
        exemplars_to_mongo(REGOLITH_MONGODB_NAME, collection_list=[mongo_coll])
    except ConnectionError:
        yield False
        return
    # Write one collection doc in file system
    fs_coll = "abstracts"
    exemplars_to_fs(fspath, collection_list=[fs_coll])
    yield repo, fs_coll, mongo_coll
    cmd = ["mongo", REGOLITH_MONGODB_NAME, "--eval", "db.dropDatabase()"]
    try:
        subprocess.check_call(cmd, cwd=repo)
    except CalledProcessError:
        print(
            f'Deleting the test database failed, insert "mongo {REGOLITH_MONGODB_NAME} --eval '
            f'"db.dropDatabase()"" into command line manually'
        )
    shut_down_fork(forked, repo)
    os.chdir(cwd)
    if not OUTPUT_FAKE_DB:
        rmtree(repo)


@pytest.fixture(scope="session")
def make_bad_db():
    """A test fixutre that creates and destroys a git repo in a temporary
    directory.
    This will yield the path to the repo.
    """
    cwd = os.getcwd()
    name = "regolith_fake_bad"
    repo = os.path.join(tempfile.gettempdir(), name)
    if os.path.exists(repo):
        rmtree(repo)
    subprocess.run(["git", "init", repo])
    os.chdir(repo)
    with open("README", "w") as f:
        f.write("testing " + name)
    with open("regolithrc.json", "w") as f:
        json.dump(
            {
                "groupname": "ERGS",
                "databases": [
                    {
                        "name": "test",
                        "url": repo,
                        "public": True,
                        "path": "db",
                        "local": True,
                    }
                ],
            },
            f,
        )
    os.mkdir("db")
    # Write collection docs
    for coll, example in deepcopy(EXEMPLARS).items():
        if isinstance(example, list):
            d = {dd["_id"]: dd for dd in example}
        else:
            d = {example["_id"]: example}
        d.update({"bad": {"_id": "bad", "bad": True}})
        if coll == "presentations":
            d.update(
                {
                    "bad_inst": {
                        "_id": "bad_inst",
                        "institution": "noinstitution",
                        "department": "nodept",
                    }
                }
            )
        dump_yaml("db/{}.yaml".format(coll), d)
    subprocess.run(["git", "add", "."])
    subprocess.run(["git", "commit", "-am", "Initial readme"])
    yield repo
    os.chdir(cwd)
    rmtree(repo)


@pytest.fixture(scope="function")
def make_fs_to_mongo_migration_db():
    """A test fixture that creates and destroys a git repo in a temporary
    directory, as well as a mongo database.
    This will yield the path to the repo.

    This specific test fixture points to a repo that contains mongo and filesystem backends with only the
    filesystem containing the exemplars. This is meant for use in migration testing.
    """
    yield from make_migration_db(True)


@pytest.fixture(scope="function")
def make_mongo_to_fs_backup_db():
    """A test fixture that creates and destroys a git repo in a temporary
    directory, as well as a mongo database.
    This will yield the path to the repo.

    This specific test fixture points to a repo that contains mongo and filesystem backends with only the
    filesystem containing the exemplars. This is meant for use in migration testing.
    """
    yield from make_migration_db(False)


def make_migration_db(fs_to_mongo_true__mongo_to_fs_false):
    cwd = os.getcwd()
    forked = False
    name = "regolith_mongo_fake"
    repo = os.path.join(tempfile.gettempdir(), name)
    if os.path.exists(repo):
        rmtree(repo)
    subprocess.run(["git", "init", repo])
    os.chdir(repo)
    with open("README", "w") as f:
        f.write("testing " + name)
    mongodbpath = os.path.join(repo, "dbs")
    os.mkdir(mongodbpath)
    fspath = os.path.join(repo, "db")
    os.mkdir(fspath)
    with open("regolithrc.json", "w") as f:
        json.dump(
            {
                "groupname": "ERGS",
                "databases": [
                    {
                        "name": ALTERNATE_REGOLITH_MONGODB_NAME,
                        "dst_url": "localhost",
                        "url": repo,
                        "path": "db",
                        "public": True,
                        "local": True,
                        "backend": "mongodb",
                    }
                ],
                "mongodbpath": mongodbpath,
            },
            f,
        )
    if os.name == "nt":
        # If on windows, the mongod command cannot be run with the fork or syslog options.
        # Instead, it is installed as a service and the exceptions that would typically be log outputs
        # are handled by the exception handlers below. In addition, the database must always be manually
        # deleted from the windows mongo instance before running a fresh test.
        cmd = ["mongo", ALTERNATE_REGOLITH_MONGODB_NAME, "--eval", "db.dropDatabase()"]
        try:
            subprocess.check_call(cmd, cwd=repo)
        except CalledProcessError:
            print(
                "Mongod likely has not been installed as a service. In order to run mongodb tests, make sure\n"
                "to install the mongodb community edition with the following link: \n"
                "https://docs.mongodb.com/manual/installation/"
            )
            yield False
            return
        cmd = ["mongostat", "--host", "localhost", "-n", "1"]
    else:
        cmd = ["mongod", "--fork", "--syslog", "--dbpath", mongodbpath]
        forked = True
    try:
        subprocess.check_call(cmd, cwd=repo)
    except CalledProcessError:
        print(
            "If on linux/mac, Mongod command failed to execute. If on windows, the status of mongo could not be\n"
            "retrieved. In order to run mongodb tests, make sure to install the mongodb community edition with\n"
            "the following link:\n"
            "https://docs.mongodb.com/manual/installation/"
        )
        yield False
        return
    if fs_to_mongo_true__mongo_to_fs_false:
        exemplars_to_fs(fspath)
    else:
        exemplars_to_mongo(ALTERNATE_REGOLITH_MONGODB_NAME)
    yield repo
    cmd = ["mongo", ALTERNATE_REGOLITH_MONGODB_NAME, "--eval", "db.dropDatabase()"]
    try:
        subprocess.check_call(cmd, cwd=repo)
    except CalledProcessError:
        print(
            f'Deleting the test database failed, insert "mongo {ALTERNATE_REGOLITH_MONGODB_NAME} --eval '
            f'"db.dropDatabase()"" into command line manually'
        )
    shut_down_fork(forked, repo)
    os.chdir(cwd)
    if not OUTPUT_FAKE_DB:
        rmtree(repo)


def shut_down_fork(forked, repo):
    if forked:
        cmd = ["mongo", "admin", "--eval", "db.shutdownServer()"]
        try:
            subprocess.check_call(cmd, cwd=repo)
        except CalledProcessError:
            print(
                'Deleting the test database failed, insert "mongo admin --eval '
                '"db.shutdownServer()"" into command line manually'
            )


def exemplars_to_fs(fspath, collection_list=None):
    exemplars_copy = deepcopy(EXEMPLARS)
    if collection_list is None:
        exemplars = exemplars_copy
    else:
        exemplars = {k: exemplars_copy[k] for k in collection_list if k in exemplars_copy}
    cwd = os.getcwd()
    os.chdir(fspath)
    for coll, example in exemplars.items():
        if isinstance(example, list):
            d = {dd["_id"]: dd for dd in example}
        else:
            d = {example["_id"]: example}
        dump_yaml("{}.yaml".format(coll), d)
    os.chdir(cwd)


def exemplars_to_mongo(mongo_db_name, collection_list=None):
    exemplars_copy = deepcopy(EXEMPLARS)
    if collection_list is None:
        exemplars = exemplars_copy
    else:
        exemplars = {k: exemplars_copy[k] for k in collection_list if k in exemplars_copy}
    client = MongoClient("localhost", serverSelectionTimeoutMS=2000)
    client.server_info()
    for col_name, example in exemplars.items():
        db = client[mongo_db_name]
        col = db[col_name]
        try:
            if isinstance(example, list):
                for doc in example:
                    doc["_id"].replace(".", "")
                col.insert_many(example)
            else:
                example["_id"].replace(".", "")
                col.insert_one(example)
        except mongo_errors.DuplicateKeyError:
            print("Duplicate key error, check exemplars for duplicates if tests fail")
        except mongo_errors.BulkWriteError:
            print("Duplicate key error, check exemplars for duplicates if tests fail")



"""Builder for Grant Reports"""

from datetime import date

import dateutil.parser as date_parser

# from habanero import Crossref
from nameparser import HumanName

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.dates import get_dates, is_current
from regolith.fsclient import _id_key
from regolith.tools import all_docs_from_collection, filter_presentations, fuzzy_retrieval


class GrantReportBuilder(LatexBuilderBase):
    """Build a proposal review from database entries"""

    btype = "grant-report"
    needed_dbs = [
        "presentations",
        "projecta",
        "people",
        "grants",
        "institutions",
        "expenses",
        "citations",
        "contacts",
    ]

    #    def __init__(self, rc):
    #        super().__init__(rc)
    #        self.needed_dbs = needed_dbs

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        for dbs in self.needed_dbs:
            gtx[dbs] = sorted(all_docs_from_collection(rc.client, dbs), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        rc = self.rc

        if not rc.grants:
            raise RuntimeError("Error: no grant specified. Please rerun specifying a grant")
        if isinstance(rc.grants, str):
            rc.grants = [rc.grants]
        if len(rc.grants) > 1:
            raise RuntimeError("Error: more than one grant specified. Please rerun with" "only a single grant.")
        grant_id = rc.grants[0]
        grant = fuzzy_retrieval(self.gtx["grants"], ["_id", "alias", "name"], grant_id)
        grant_dates = get_dates(grant)

        # Convert Date Strings to Datetime Objects
        if rc.from_date:
            rp_start_date = date_parser.parse(rc.from_date).date()
        else:
            rp_start_date = grant_dates.get("begin_date")
            print(
                f"INFO: no begin-date specified.  running report from the beginning "
                f"of the grant period ({rp_start_date})"
            )
        if rc.to_date:
            rp_end_date = date_parser.parse(rc.to_date).date()
        else:
            rp_end_date = min([date.today(), grant_dates.get("end_date")])
            print(
                "INFO: no end-date specified for the reporting period.  Running "
                "report up to the earlier of the end of the grant, or today "
                f"({rp_end_date})."
            )
        report_dates = {"begin_date": rp_start_date, "end_date": rp_end_date}
        print(
            f"INFO: generating report for grant {grant_id} for the period"
            f"from {rp_start_date} to {rp_end_date})"
        )

        # Get prum associated to grant and active during reporting period
        #        institutions_coll = [inst for inst in self.gtx["institutions"]]
        institutions_coll = self.gtx["institutions"]
        grant_prums = [
            prum
            for prum in self.gtx["projecta"]
            if grant_id in prum.get("grants", []) and "checklist" not in prum.get("deliverable").get("scope")
        ]
        #        for prum in self.gtx['projecta']:
        #            if grant_name in prum['grants']:
        #                begin_date = get_dates(prum).get('begin_date')
        #                due_date = get_due_date(prum['deliverable'])
        #                # if projectum was finished during reporting period or is still current
        #                # some projectum don't have an "end date", but all projecta have a deliverable
        #                # due_date
        #                if (rp_start_date <= due_date <= rp_end_date and \
        #                prum['status'] is "finished") or is_current(prum):
        #                   grant_prums.append(prum)
        # Get people associated with grant

        grant_prums_finished_this_period = [
            prum for prum in grant_prums if is_current(report_dates, get_dates(prum).get("end_date"))
        ]
        grant_prum_leads = list(set([prum["lead"] for prum in grant_prums]))
        grant_prum_collaborators = list(
            set([collab for prum in grant_prums for collab in prum.get("collaborators", [])])
        )
        list(set([grp_mbr for prum in grant_prums for grp_mbr in prum.get("group_members", [])]))
        grant_people = grant_prum_leads
        # Accomplishments
        major_activities = []
        significant_results = []
        for prum in grant_prums:
            if prum["status"] == "finished":
                continue
            else:
                major_activities.append(prum)
        for prum in grant_prums_finished_this_period:
            significant_results.append(prum)

        # Opportunities for Training and Professional Development
        training_and_professional_development = []
        # presentations
        for id in grant_people:
            training_and_professional_development.extend(
                filter_presentations(
                    self.gtx["people"],
                    self.gtx["presentations"],
                    institutions_coll,
                    id,
                    types=["all"],
                    since=rp_start_date,
                    before=rp_end_date,
                    statuses=["accepted"],
                )
            )
        # thesis defendings
        # how do i access people.yml in rg-db-public vs the people.yml file in rg-db-group?
        #        defended_theses = []
        #        for id in grant_people:
        #            for prsn in self.gtx['people']:
        #                if prsn["_id"] != id:
        #                    continue
        #                else:
        #                    person = prsn
        #            for education in person['education']:
        #                edu_dates = get_dates(education)
        #                if 'phd' in education['degree'].lower() and 'columbia' in \
        #                education['institution'].lower() and rp_start_date.year <= edu_dates.get(\
        #                'end_date', edu_dates['date']).year <= rp_end_date.year:
        #                    defended_theses.append(id)

        # Products
        # need rg-db-public's citation.yml
        #        publications = filter_publications(self.gtx["citations"],
        #                                           set(grant_people),
        #                                           since=rp_start_date,
        #                                          before=rp_end_date)
        publications = [publ for publ in self.gtx["citations"] if grant_id in publ.get("grant", "")]

        for publ in publications:
            formatted_authors = [HumanName(name).full_name for name in publ.get("authors", [])]
            publ["authors"] = formatted_authors
        # Participants/Organizations
        participants = []
        for person in self.gtx["people"]:
            months_on_grant, months_left = self.months_on(grant_id, person, rp_start_date, rp_end_date)
            if months_on_grant > 0:
                participants.append(
                    {
                        "name": person.get("name"),
                        "email": person.get("email"),
                        "position": person.get("position"),
                        "months_on_grant": int(round(months_on_grant, 0)),
                    }
                )

        collaborators = {}
        missing_contacts = []
        for id in grant_prum_collaborators:
            for contact in self.gtx["contacts"]:
                if contact["_id"] == id:
                    name = contact.get("name")
                    aka = contact.get("aka")
                    institution_id = contact.get("institution")
                    institution = fuzzy_retrieval(institutions_coll, ["name", "aka", "_id"], institution_id)
                    if institution:
                        inst_name = institution.get("name")
                    else:
                        print(f"WARNING: institution {institution_id} not found " f"in institutions collection")
                        inst_name = institution_id
                    collaborators[id] = {"aka": aka, "name": name, "institution": inst_name}
        missing_contacts = [id for id in grant_prum_collaborators if not collaborators.get(id)]
        missing_contacts = list(set(missing_contacts))
        for person_id in missing_contacts:
            print(f"WARNING: contact {person_id} not found in contacts collection")

        # Impacts
        begin_date_str = rp_start_date.isoformat()
        end_date_str = rp_end_date.isoformat()
        self.render(
            "grantreport.txt",
            f"{grant_id}_report_{begin_date_str}_{end_date_str}.txt",
            begin_date=begin_date_str,
            newline="\n",
            end_date=end_date_str,
            majorActivities=major_activities,
            significantResults=significant_results,
            trainingAndProfessionalDevelopment=training_and_professional_development,
            #            defendedTheses=defended_theses,
            products=publications,
            grantPeople=grant_people,
            participants=participants,
            collaborators=collaborators,
            hline="------------------------------------------------------------------------------",
        )

    def months_on(self, grant, person, since=date(1970, 1, 1), before=date.today()):
        #    print('Looking at months on grant {} in period since {} until {}'.format(
        #        grant, since, before), )
        total_months = 0
        appts = person.get("appointments")
        if appts:
            months = 0
            for k1, v1 in appts.items():
                if grant in v1.get("grant"):
                    appt_dates = get_dates(v1)
                    overlap_start = max([appt_dates.get("begin_date"), since])
                    overlap_end = min([appt_dates.get("end_date"), before])
                    if overlap_end >= overlap_start:
                        months = months + (overlap_end - overlap_start).days * v1.get("loading") / 30.4
                    # appt_startdate = dates.get('begin_date')
                    # appt_enddate = dates.get('end_date')
                    # loading = v1.get('loading')
                    # if appt_enddate >= since and appt_startdate <= before:
                    #    months = months + (
                    #            app_enddate - appt_startdate).days * loading / 30.4
                    # elif startdate >= since and enddate > before:
                    #    months = months + (
                    #            before - startdate).days * loading / 30.4
                    # elif startdate < since and enddate <= before:
                    #    months = months + (
                    #            enddate - since).days * loading / 30.4
                    # elif startdate < since and enddate > before:
                    #    months = months + (before - since).days * loading / 30.4
            if months > 0:
                total_months = total_months + months
        months_left = before - date.today()
        return total_months, months_left



import operator

from gooey import GooeyParser

from regolith.dates import get_dates
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import all_docs_from_collection, dereference_institution, get_person_contact, get_pi_id

TARGET_COLL = "presentations"
HELPER_TARGET = "l_abstract"


def subparser(subpi):
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        int_kwargs["widget"] = "IntegerField"
        int_kwargs["gooey_options"] = {"min": 2000, "max": 2100}

    subpi.add_argument("-a", "--author", help="Filter abstracts or this author ID (e.g., sbillinge).")
    subpi.add_argument("-y", "--year", help="Get presentations since this year")
    subpi.add_argument(
        "-l",
        "--loc-inst",
        help="location of presentation, either a fragment of an institution, "
        "country, city, state, or university. If an institution is entered,"
        "the search will be for seminars or colloquiums, otherwise the "
        "search will be for all other meetings",
    )
    subpi.add_argument(
        "-t", "--title", help="fragment of the title of the abstract or talk to use to " "filter presentations"
    )
    return subpi


class AbstractListerHelper(SoutHelperBase):
    """Helper for finding and listing abstracts from the presentations.yml file"""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "people", "contacts", "institutions"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc

        presentations = self.gtx["presentations"]
        institutions = self.gtx["institutions"]
        SEMINAR_TYPES = ["seminar", "colloquium"]
        filtered_title, filtered_authors, filtered_years, filtered_inst, filtered_loc = ([] for i in range(5))

        if (not rc.author) and (not rc.year) and (not rc.loc_inst) and (not rc.title):
            print("-------------------------------------------")
            print("please rerun specifying at least one filter")
            print("-------------------------------------------")
            return

        if rc.title:
            filtered_title = [
                presentation
                for presentation in presentations
                if rc.title.casefold() in presentation.get("title").casefold()
            ]
        else:
            filtered_title = [presentation for presentation in presentations]

        if rc.author:
            filtered_authors = [
                presentation for presentation in filtered_title if rc.author in presentation.get("authors")
            ]
        else:
            filtered_authors = filtered_title

        if rc.year:
            filtered_years = [
                presentation
                for presentation in filtered_authors
                if get_dates(presentation)
                .get("date", get_dates(presentation).get("end_date", get_dates(presentation).get("begin_date")))
                .year
                == int(rc.year)
            ]
        else:
            filtered_years = filtered_authors

        if rc.loc_inst:
            filtered_inst = [
                presentation
                for presentation in filtered_years
                if presentation.get("type") in SEMINAR_TYPES
                and rc.loc_inst.casefold() in presentation.get("institution", "").casefold()
            ]
            filtered_loc = [
                presentation
                for presentation in filtered_years
                if rc.loc_inst.casefold() in presentation.get("location", "").casefold()
            ]
        else:
            filtered_inst = filtered_years
            filtered_loc = filtered_years

        filtered_presentations_by_args = [filtered_inst, filtered_loc]

        nonempty_filtered_presentations_by_args = [
            filtered_presentations
            for filtered_presentations in filtered_presentations_by_args
            if filtered_presentations
        ]
        filtered_presentations = [
            talk
            for presentations in nonempty_filtered_presentations_by_args
            for talk in presentations
            if all(talk in presentations for presentations in nonempty_filtered_presentations_by_args)
        ]
        for talk in filtered_presentations:
            talk.update(
                {
                    "meeting_date": get_dates(talk).get(
                        "date", get_dates(talk).get("end_date", get_dates(talk).get("begin_date"))
                    )
                }
            )
        filtered_presentations_sorted = sorted(filtered_presentations, key=operator.itemgetter("meeting_date"))
        flat_filtered_presentations = list({talk["_id"]: talk for talk in filtered_presentations_sorted}.values())
        for presentation in flat_filtered_presentations:
            if presentation.get("type") in SEMINAR_TYPES:
                dereference_institution(presentation, institutions)
                meeting_info = (
                    f"{presentation.get('type').title()} {presentation.get('department')}, "
                    f"{presentation.get('institution')}"
                )
            else:
                meeting_info = f"{presentation.get('meeting_name')}, {presentation.get('location')}"
            print("\n---------------------------------------")
            print(f"{presentation.get('meeting_date').isoformat()} - {meeting_info}")
            print("---------------------------------------")
            print(f"Title: {presentation.get('title')}\n")
            author_list = [
                (
                    author
                    if not get_person_contact(author, self.gtx["people"], self.gtx["contacts"])
                    else get_person_contact(author, self.gtx["people"], self.gtx["contacts"]).get("name")
                )
                for author in presentation.get("authors")
            ]
            print(", ".join(author_list))
            print(f"\nAbstract: {presentation.get('abstract')}")
        return



"""
Helper to add expenses.
"""

import datetime as dt

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.schemas import alloweds
from regolith.tools import all_docs_from_collection, get_pi_id

TARGET_COLL = "expenses"

EXPENSES_STATI = alloweds.get("EXPENSES_STATI")
EXPENSES_TYPES = alloweds.get("EXPENSES_TYPES")


def expense_constructor(key, begin_date, end_date, rc):
    """
    constructs a document with default fields for an expense

    Parameters
    ----------
    key str
      The key for ('_id') for this entry in the expenses collection
    begin_date datetime.date
      The date of the start of the travel/expense
    end_date datetime.date
      The date of the end of the travel/expense
    rc runcontrol object
      The runcontrol object.  Should contain:
         required:
         rc.business (bool) True if a business expense, false if a travel one.
         rc.grants (list) The list of grants that will be charged
         rc.purpose (str) The purpose of the trip/expense
         rc.payee (str) id of the payee
         rc.where (str) where submitted/reimbursed (e.g., which institution/person
            it was submitted to or which bank account it was reimbursed to
         required if rc.business:
         rc.amount (float) The amount of the expense if it is a business expense
         optional:
         rc.notes (list of strings) notes to add if there are any
         rc.status (str) the status of the expense (defaults to unsubmitted)

    Returns
    -------
    The constructed expense document

    """
    pdoc = {}
    pdoc.update(
        {
            "_id": key,
            "begin_date": begin_date,
            "end_date": end_date,
        }
    )

    if rc.business:
        expense_type = "business"
    else:
        expense_type = "travel"
    pdoc.update({"expense_type": expense_type})

    percentages = [round(100 / len(rc.grants), 2) for i in rc.grants]

    pdoc.update({"grant_percentages": percentages, "grants": rc.grants})

    if expense_type == "travel":
        pdoc.update(
            {
                "itemized_expenses": [
                    {
                        "date": begin_date,
                        "purpose": "registration",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "home to airport",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "flights",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "airport to hotel",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "hotel",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "hotel to airport",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "airport to home",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                    {
                        "date": begin_date,
                        "purpose": "meals",
                        "unsegregated_expense": 0,
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    },
                ]
            }
        )
    else:
        pdoc.update(
            {
                "itemized_expenses": [
                    {
                        "date": begin_date,
                        "purpose": rc.purpose,
                        "unsegregated_expense": float(rc.amount),
                        "segregated_expense": 0,
                        "currency": "USD",
                        "notes": [""],
                    }
                ]
            }
        )
    pdoc.update(
        {
            "notes": rc.notes,
            "overall_purpose": rc.purpose,
            "payee": rc.payee,
        }
    )
    if rc.status == "submitted":
        submission_date = dt.date.today()

    else:
        submission_date = "tbd"

    pdoc.update(
        {
            "reimbursements": [
                {
                    "amount": 0,
                    "date": "tbd",
                    "submission_date": submission_date,
                    "where": rc.where,
                }
            ]
        }
    )
    pdoc.update({"status": rc.status})
    return pdoc


def subparser(subpi):
    amount_gooey_kwargs, notes_gooey_kwargs, date_gooey_kwargs = {}, {}, {}
    if isinstance(subpi, GooeyParser):
        amount_gooey_kwargs["widget"] = "DecimalField"
        amount_gooey_kwargs["gooey_options"] = {"min": 0.00, "max": 1000000.00, "increment": 10.00, "precision": 2}
        notes_gooey_kwargs["widget"] = "Textarea"
        date_gooey_kwargs["widget"] = "DateChooser"

    subpi.add_argument("name", help="A short name for the expense", default=None)
    subpi.add_argument(
        "purpose", help="A short description of the business " "purpose of the expense", default=None
    )

    subpi.add_argument(
        "-b",
        "--business",
        action="store_true",
        help="Is the expense type business? If not specified, defaults to travel",
    )
    subpi.add_argument(
        "-a", "--amount", help="expense amount. required if a business " "expense.", **amount_gooey_kwargs
    )
    subpi.add_argument(
        "-d",
        "--begin-date",
        help="Input begin date for this expense. " "Defaults to today's date",
        **date_gooey_kwargs,
    )
    subpi.add_argument(
        "-e,",
        "--end-date",
        help="Input end date for this expense. " "Defaults to today's date",
        **date_gooey_kwargs,
    )
    subpi.add_argument(
        "-g",
        "--grants",
        nargs="+",
        help="grant, or list of grants that cover this expense. Defaults to tbd",
        default="tbd",
    )
    subpi.add_argument(
        "-s",
        "--status",
        choices=EXPENSES_STATI,
        help=f"status, from {EXPENSES_STATI}. Default is unsubmitted",
        default="unsubmitted",
    )
    subpi.add_argument("-w", "--where", help="Where the expense has been submitted.", default="")
    subpi.add_argument(
        "-n",
        "--notes",
        nargs="+",
        help="List of notes for the expense. Defaults to empty list",
        default=[],
        **notes_gooey_kwargs,
    )
    subpi.add_argument("-y", "--payee", help="payee of the expense. defaults to rc.default_user_id")
    # Do not delete --database arg
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    return subpi


class ExpenseAdderHelper(DbHelperBase):
    btype = "a_expense"
    needed_colls = [f"{TARGET_COLL}", "people", "groups"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        if not rc.payee:
            if rc.default_user_id:
                rc.payee = rc.default_user_id
            else:
                raise RuntimeError(
                    " No default_user_id set.  Please specify this "
                    "either in the ~/.conf/regolith/user.json or in"
                    " regolithrc.json"
                )
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        # dates
        if rc.begin_date:
            begin_date = date_parser.parse(rc.begin_date).date()
        else:
            begin_date = dt.date.today()
        if rc.end_date:
            end_date = date_parser.parse(rc.end_date).date()
        else:
            end_date = dt.date.today()
        key = (
            f"{str(begin_date.year)[2:]}{str(begin_date.strftime('%m'))}"
            f"{rc.payee[0:2]}_{''.join(rc.name.casefold().split()).strip()}"
        )
        coll = self.gtx[rc.coll]
        pdocl = list(filter(lambda doc: doc["_id"] == key, coll))
        if len(pdocl) > 0:
            raise RuntimeError("This entry appears to already exist in the collection")
        else:
            pdoc = {}

        pdoc = expense_constructor(key, begin_date, end_date, rc)

        rc.client.insert_one(rc.database, rc.coll, pdoc)

        print(f"{key} has been added in {TARGET_COLL}")

        return



"""Flask app for looking at information in regolith."""

import json
import os
import tempfile
import traceback

from flask import Flask, abort, redirect, render_template, request

from regolith.chained_db import _convert_to_dict
from regolith.schemas import validate

app = Flask("regolith")


@app.route("/", methods=["GET", "POST"])
def root():
    rc = app.rc
    if request.method == "POST":
        form = request.form
        return redirect("/db/{dbname}/coll/{collname}".format(**form))
    return render_template("index.html", rc=rc)


def shutdown_server():
    func = request.environ.get("werkzeug.server.shutdown")
    if func is None:
        raise RuntimeError("Not running with the Werkzeug Server")
    func()


@app.route("/shutdown", methods=["GET", "POST"])
def shutdown():
    shutdown_server()
    return "Regolith server shutting down...\n"


@app.route("/db/<dbname>/coll/<collname>", methods=["GET", "POST"])
def collection_page(dbname, collname):
    rc = app.rc
    try:
        coll = rc.client[dbname][collname]
    except (KeyError, AttributeError):
        abort(404)
    status = status_id = None
    if request.method == "POST":
        form = request.form
        if "shutdown" in form:
            return shutdown()
        elif "cancel" in form:
            body = json.loads(form["body"].strip())
            status = "canceled"
            status_id = str(body["_id"])
        elif "save" in form:
            try:
                body = json.loads(form["body"].strip())
            except Exception:
                td = tempfile.TemporaryDirectory()
                n = os.path.join(td.name, "regolith.txt")
                print("Error in json parsing writing text file to {}. " "Please try again.".format(n))
                with open(n, "w", encoding="utf-8") as f:
                    f.write(form["body"])
                traceback.print_exc()
                raise
            tv, errors = validate(dbname, body, rc.schemas)
            if not tv:
                td = tempfile.TemporaryDirectory()
                n = os.path.join(td.name, "regolith.txt")
                with open(n, "w", encoding="utf-8") as f:
                    f.write(form["body"])
                raise ValueError(
                    "Error while validating the record,"
                    " writing text file to {}. "
                    "Please try again.\n\n"
                    "Your errors were\n"
                    "------------------"
                    "{}".format(n, errors)
                )

            rc.client.update_one(dbname, collname, {"_id": body["_id"]}, body)
            status = "saved ✓"
            status_id = str(body["_id"])
        elif "add" in form:
            try:
                body = json.loads(form["body"].strip())
                print(body)
            except Exception:
                td = tempfile.TemporaryDirectory()
                n = os.path.join(td.name, "regolith.txt")
                print("Error in json parsing writing text file to {}. " "Please try again.".format(n))
                with open(n, encoding="utf-8") as f:
                    f.write(form["body"])
                traceback.print_exc()
                raise
            tv, errors = validate(dbname, body, rc.schemas)
            if not tv:
                td = tempfile.TemporaryDirectory()
                n = os.path.join(td.name, "regolith.txt")
                with open(n, encoding="utf-8") as f:
                    f.write(form["body"])
                raise ValueError(
                    "Error while validating the record,"
                    " writing text file to {}. "
                    "Please try again.\n\n"
                    "Your errors were\n"
                    "------------------"
                    "{}".format(n, errors)
                )
            try:
                rc.client.insert_one(dbname, collname, body)
            except Exception:
                traceback.print_exc()
                raise
            status = "added ✓"
            status_id = str(body["_id"])
        elif "delete" in form:
            body = json.loads(form["body"].strip())
            rc.client.delete_one(dbname, collname, body)
    return render_template(
        "collection.html",
        rc=rc,
        dbname=dbname,
        len=len,
        str=str,
        status=status,
        status_id=status_id,
        collname=collname,
        coll=coll,
        json=json,
        min=min,
        conv=_convert_to_dict,
    )



"""Helper for managing appointments.

- Returns members with gap in appointments
- Returns members supported on an outdated grant
- Returns members supported on a depleted grant
- Suggests appointments to make for these members
- Suggests new appointments
"""

from datetime import date, timedelta

import matplotlib.pyplot as plt
import numpy
from dateutil import parser as date_parser
from dateutil.relativedelta import relativedelta
from gooey import GooeyParser

from regolith.dates import get_dates

# from regolith.schemas import APPOINTMENTS_TYPES
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import (
    all_docs_from_collection,
    collect_appts,
    fuzzy_retrieval,
    get_pi_id,
    grant_burn,
    group_member_employment_start_end,
    is_fully_appointed,
    merge_collections_superior,
)

TARGET_COLL = "people"
HELPER_TARGET = "makeappointments"
BLACKLIST = [
    "ta",
    "physmatch",
    "chemmatch",
    "bridge16",
    "collgf",
    "afgrf14",
    "summer@seas",
    "frap",
    "startup",
    "they_pay",
]
MONTHLY_COST_QUANTUM = 3262

_future_grant = {
    "_id": "_future_grant",
    "account": "n/a",
    "activity": 0,
    "admin": "tbd",
    "alias": "future_grant",
    "amount": 0,
    "awardnr": "tbd",
    "budget": [
        {
            "begin_date": "2020-05-01",
            "end_date": "2099-12-31",
            "amount": 0,
            "student_months": 0,
            "postdoc_months": 0,
            "ss_months": 0,
        }
    ],
}


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"
    else:
        subpi.add_argument(
            "run",
            help="Run the helper"
            'The grant "future_grant" is available internally '
            "to assign people to for making projections.  It "
            "will be plotted to show when you need new funding "
            "by and how much.",
        )
    subpi.add_argument(
        "-d",
        "--projection-from-date",
        help="the date from which projections into the future " "will be calculated",
        **date_kwargs,
    )
    subpi.add_argument("--no-plot", action="store_true", help="suppress plotting feature")
    subpi.add_argument(
        "--no-gui", action="store_true", help="suppress interactive matplotlib GUI (used for " "running tests)"
    )
    subpi.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Plot all non-blacklisted grants.  If not set, grants "
        "that ended more than 2 years ago won't be plotted",
    )
    # Do not delete --database arg
    subpi.add_argument(
        "--database", help="The database that will be updated. Defaults to first database in regolithrc.json"
    )

    return subpi


def plotter(datearray, student=None, pd=None, ss=None, title=None):
    fig, ax = plt.subplots()
    sta = numpy.array(student) / 30.5
    pda = numpy.array(pd) / 30.5
    ssa = numpy.array(ss) / 30.5
    if student:
        ax.plot_date(datearray, sta, fmt=",-", label="student months")
    if pd:
        ax.plot_date(datearray, pda, fmt=",-", label="postdoc months")
    if ss:
        ax.plot_date(datearray, ssa, fmt=",-", label="ss months")
    if student and pd:
        ax.plot_date(datearray, sta + pda, fmt=",-", label="student+postdoc days")
    ax.set_xlabel("date")
    ax.set_ylabel("budget months remaining")
    ax.set_title(title)
    ax.legend(loc="best")
    fig.autofmt_xdate()
    return fig, ax, "plotting mode is on"


class MakeAppointmentsHelper(SoutHelperBase):
    """Helper for managing appointments on grants and studying the burn of grants over time."""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "grants", "proposals"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        try:
            if not rc.database:
                rc.database = rc.databases[0]["name"]
        except Exception:
            pass
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        outdated, depleted, underspent, overspent = [], [], [], []
        people = list(self.gtx["people"])
        all_appts = collect_appts(people, filter_key="type", filter_value="gra")
        all_appts.extend(collect_appts(people, filter_key="type", filter_value="ss"))
        all_appts.extend(collect_appts(people, filter_key="type", filter_value="pd"))
        if rc.projection_from_date:
            projection_from_date = date_parser.parse(rc.projection_from_date).date()
        else:
            projection_from_date = date.today()

        # collecting amounts and time interval for all grants
        _future_grant["begin_date"] = projection_from_date
        _future_grant["end_date"] = projection_from_date + timedelta(days=2190)
        _future_grant["budget"][0]["begin_date"] = projection_from_date
        _future_grant["budget"][0]["end_date"] = projection_from_date + timedelta(days=2190)
        _future_grant["burn"] = grant_burn(_future_grant, all_appts)
        all_grants = merge_collections_superior(self.gtx["proposals"], self.gtx["grants"], "proposal_id")
        all_grants.append(_future_grant)
        most_grants_id = [grant for grant in all_grants if grant.get("_id") not in BLACKLIST]
        most_grants = [grant for grant in most_grants_id if grant.get("alias") not in BLACKLIST]
        collecting_grants_with_appts = []
        for person in self.gtx["people"]:
            appts = collect_appts([person], filter_key="type", filter_value="gra")
            appts.extend(collect_appts([person], filter_key="type", filter_value="ss"))
            appts.extend(collect_appts([person], filter_key="type", filter_value="pd"))
            if len(appts) > 0:
                person.update({"appts": appts})
                collecting_grants_with_appts.extend([appt.get("grant") for appt in appts])
        grants_with_appts = list(set(collecting_grants_with_appts))
        appointed_grants = [
            grant
            for grant in most_grants
            if grant.get("_id") in grants_with_appts or grant.get("alias") in grants_with_appts
        ]
        grants_end, grants_begin = None, None
        for grant in appointed_grants:
            grant["burn"] = grant_burn(grant, all_appts)
            grant_begin = get_dates(grant)["begin_date"]
            grant_end = get_dates(grant)["end_date"]
            grant.update({"begin_date": grant_begin, "end_date": grant_end})
            if not grants_begin or grant_begin < grants_begin:
                grants_begin = grant_begin
            if not grants_end or grant_end > grants_end:
                grants_end = grant_end

        # checking appointments
        cum_months_to_cover = 0
        for person in self.gtx["people"]:
            if not person.get("appts"):
                continue
            appts = person.get("appts")
            person_dates = group_member_employment_start_end(person, "bg")
            emps = [person_date for person_date in person_dates if not person_date.get("permanent")]
            emps.sort(key=lambda x: x.get("end_date", 0))
            is_fully_appointed(
                person,
                min(get_dates(appt)["begin_date"] for appt in appts),
                max(get_dates(appt)["end_date"] for appt in appts),
            )
            for appt in appts:
                if appt.get("grant") in BLACKLIST:
                    continue
                this_grant = fuzzy_retrieval(appointed_grants, ["_id", "alias"], appt.get("grant"))
                if not this_grant:
                    raise RuntimeError(
                        "    grant: {}, person: {}, appointment: {}, grant not found in grants database".format(
                            appt.get("grant"), person.get("_id"), appt.get("_id")
                        )
                    )
                appt_begin, appt_end = get_dates(appt)["begin_date"], get_dates(appt)["end_date"]
                outdated_period, depleted_period = False, False
                for x in range((appt_end - appt_begin).days + 1):
                    day = appt_begin + relativedelta(days=x)
                    if not outdated_period:
                        if not this_grant.get("burn"):
                            print(this_grant.get("_id"))
                        if not this_grant["burn"].get(day):
                            outdated_period = True
                            outdated.append(
                                "    person: {}, appointment: {}, grant: {},\n"
                                "            from {} until {}".format(
                                    person.get("_id"),
                                    appt.get("_id"),
                                    appt.get("grant"),
                                    (
                                        str(day)
                                        if day < this_grant["begin_date"]
                                        else this_grant["end_date"] + relativedelta(days=1)
                                    ),
                                    (
                                        str(min(appt_end, this_grant["begin_date"]))
                                        if day < this_grant["begin_date"]
                                        else str(day)
                                    ),
                                )
                            )
                    else:
                        if this_grant["burn"].get(day):
                            outdated_period = False
                    if not (depleted_period or outdated_period):
                        day_burn, this_burn = 0, this_grant["burn"]
                        if appt.get("type") == "gra":
                            day_burn = this_burn[day]["student_days"]
                        elif appt.get("type") == "pd":
                            day_burn = this_burn[day]["postdoc_days"]
                        elif appt.get("type") == "ss":
                            day_burn = this_burn[day]["ss_days"]
                        if day_burn < -5:
                            # FIXME change to display depleted until next >-5 amt instead of appt_end
                            depleted.append(
                                "    person: {}, appointment: {}, grant: {},\n"
                                "            from {} until {}".format(
                                    person["_id"], appt["_id"], appt.get("grant"), str(day), str(appt_end)
                                )
                            )
                            depleted_period = True

        # setup for plotting grants
        datearray, cum_student, cum_pd, cum_ss = [], None, None, None
        if not rc.no_plot:
            for x in range((grants_end - grants_begin).days + 1):
                datearray.append(grants_begin + relativedelta(days=x))
            cum_student, cum_pd, cum_ss = [0.0] * len(datearray), [0.0] * len(datearray), [0.0] * len(datearray)
        plots = []

        # calculating grant surplus and deficit
        cum_underspend = 0
        for grant in appointed_grants:
            tracking = [balance for balance in grant.get("tracking", []) if balance]
            # if all_grants[grant]:
            #     tracking = [balance for balance in all_grants[grant].get('tracking',[]) if balance]
            # else:
            #     tracking = []
            if len(tracking) > 0:
                tracking.sort(key=lambda x: x[0])
                recent_balance = tracking[-1]
                recent_balance[1] = recent_balance[1] / MONTHLY_COST_QUANTUM
            else:
                recent_balance = [projection_from_date, 0]
            budget_begin = min(get_dates(period)["begin_date"] for period in grant.get("budget"))
            budget_end = max(get_dates(period)["end_date"] for period in grant.get("budget"))
            if grant["begin_date"] != budget_begin:
                raise RuntimeError(
                    f"grant {grant.get('alias')} does not have a correct budget begin date. "
                    f"grant begin: {grant['begin_date']} budget begin: {budget_begin}"
                )
            elif grant["end_date"] != budget_end:
                raise RuntimeError(
                    f"grant {grant.get('alias')} does not have a correct budget end date."
                    f" grant end: {grant['end_date']} budget end: {budget_end}"
                )
            days_to_go = (grant["end_date"] - projection_from_date).days
            this_burn = grant["burn"]
            end_amount = (
                this_burn.get(grant["end_date"])["student_days"]
                + this_burn.get(grant["end_date"])["ss_days"]
                + this_burn.get(grant["end_date"])["postdoc_days"]
            )
            if end_amount > 15.25:
                underspent.append(
                    (
                        grant["end_date"],
                        grant.get("alias"),
                        round(end_amount / 30.5, 2),
                        round(end_amount / days_to_go, 2),
                    )
                )
                cum_underspend += end_amount
            elif end_amount < -30.5:
                overspent.append(
                    "    end: {}, grant: {}, overspend amount: {} months".format(
                        str(grant["end_date"]), grant.get("alias"), round(end_amount / 30.5, 2)
                    )
                )
            # values for individual and cumulative grant burn plots
            if not rc.no_plot:
                grant_dates = [
                    grant["begin_date"] + relativedelta(days=x)
                    for x in range((grant["end_date"] - grant["begin_date"]).days + 1)
                ]
                this_student, this_pd, this_ss = (
                    [0.0] * len(grant_dates),
                    [0.0] * len(grant_dates),
                    [0.0] * len(grant_dates),
                )
                counter = 0
                for x in range(len(datearray)):
                    day_burn = this_burn.get(datearray[x])
                    if day_burn:
                        this_student[counter] = day_burn["student_days"]
                        this_pd[counter] = day_burn["postdoc_days"]
                        this_ss[counter] = day_burn["ss_days"]
                        cum_student[x] += day_burn["student_days"]
                        cum_pd[x] += day_burn["postdoc_days"]
                        cum_ss[x] += day_burn["ss_days"]
                        counter += 1
                if not rc.verbose:
                    if max(grant_dates) >= projection_from_date - timedelta(days=730):
                        plots.append(
                            plotter(
                                grant_dates, student=this_student, pd=this_pd, ss=this_ss, title=grant.get("alias")
                            )[0]
                        )
                else:
                    plots.append(
                        plotter(
                            grant_dates, student=this_student, pd=this_pd, ss=this_ss, title=grant.get("alias")
                        )[0]
                    )

        if outdated:
            outdated.sort(key=lambda mess: mess[-10:])
            print("appointments on outdated grants:")
            for appt in outdated:
                print(appt)
        if depleted:
            depleted.sort(key=lambda mess: mess[-10:])
            print("appointments on depleted grants:")
            for appt in depleted:
                print(appt)
        if underspent:
            underspent.sort(key=lambda x: x[0])
            print("underspent grants:")
            for grant_info in underspent:
                print(
                    f"    {grant_info[1]}: end: {grant_info[0]}\n"
                    f"      projected underspend: {grant_info[2]} months, "
                    f"balance as of {recent_balance[0]}: {recent_balance[1]}\n"
                    f"      required ss+gra burn: {grant_info[3]}"
                )
            print(
                f"cumulative underspend = {round(cum_underspend/30.5, 2)} months, "
                f"cumulative months to support = {round(cum_months_to_cover, 2)}"
            )
        if overspent:
            print("overspent grants:")
            for grant in overspent:
                print(grant)

        if not rc.no_plot:
            for plot in plots:
                if not rc.no_gui:
                    plt.show()
            cum_plot, cum_ax, outp = plotter(
                datearray, student=cum_student, pd=cum_pd, ss=cum_ss, title="Cumulative burn"
            )
            if not rc.no_gui:
                plt.show()

            print(outp)

        return



"""Helper for listing upcoming (and past) projectum milestones.

Projecta are small bite-sized project quanta that typically will result in
one manuscript.
"""

import datetime as dt

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.schemas import (
    PROJECTUM_ACTIVE_STATI,
    PROJECTUM_CANCELLED_STATI,
    PROJECTUM_FINISHED_STATI,
    PROJECTUM_PAUSED_STATI,
)
from regolith.tools import all_docs_from_collection, collection_str, get_pi_id, key_value_pair_filter

TARGET_COLL = "projecta"
HELPER_TARGET = "l_projecta"
INACTIVE_STATI = PROJECTUM_PAUSED_STATI + PROJECTUM_CANCELLED_STATI


def subparser(subpi):
    date_kwargs = {}
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"
        int_kwargs["widget"] = "IntegerField"

    subpi.add_argument(
        "-l", "--lead", help="Filter milestones for this project lead, " "specified by id, e.g., sbillinge"
    )
    subpi.add_argument(
        "-p",
        "--person",
        help="Filter milestones for this person whether lead or " "not, specified by id, e.g., sbillinge",
    )
    subpi.add_argument("-g", "--grant", help="Filter projecta by a grant ID")
    subpi.add_argument(
        "-c",
        "--current",
        action="store_true",
        help="Lists only active projecta. If not specified, will be active and paused but not cancelled",
    )
    subpi.add_argument("-v", "--verbose", action="store_true", help="increase verbosity of output")
    subpi.add_argument("--grp_by_lead", action="store_true", help="Lists all projecta by their lead")
    subpi.add_argument(
        "-o",
        "--orphans",
        action="store_true",
        help="Find all orphans: prums that are assigned to 'tbd' or have a " "non active person as lead",
    )
    subpi.add_argument(
        "--all",
        action="store_true",
        help=f"Lists all projecta including those with statuses "
        f"in {*PROJECTUM_CANCELLED_STATI, } that are excluded by default",
    )
    subpi.add_argument(
        "-e",
        "--ended",
        action="store_true",
        help="Lists projecta that have ended. Use the -r flag to specify "
        "how many days prior to today to search over for finished"
        "prums (default is 7 days). "
        "You may also use -d to change date used to search back"
        "from (mostly used for testing)",
    )
    subpi.add_argument(
        "-r",
        "--range",
        help="date range back from DATE, in days, to search over "
        "for finished prums."
        "If no range is specified, search range will be 7 days",
    )
    subpi.add_argument("-f", "--filter", nargs="+", help="Search this collection by giving key element pairs")
    subpi.add_argument(
        "-k",
        "--keys",
        nargs="+",
        help="Specify what keys to return values from when running --filter. "
        "If no argument is given the default is just the id.",
    )
    subpi.add_argument(
        "-d",
        "--date",
        help="projecta with end_date within RANGE before this date will be listed. "
        "The default is today. Some projecta don't have an end date and won't appear in a search. "
        "mostly used for testing.",
        **date_kwargs,
    )
    return subpi


class ProjectaListerHelper(SoutHelperBase):
    """Helper for listing upcoming (and past) projectum milestones.

    Projecta are small bite-sized project quanta that typically will result in
    one manuscript.
    """

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "people"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        if rc.filter:
            collection = key_value_pair_filter(self.gtx["projecta"], rc.filter)
        else:
            collection = self.gtx["projecta"]

        if (
            (not rc.lead)
            and (not rc.person)
            and (not rc.ended)
            and (not rc.grant)
            and (not rc.verbose)
            and (not rc.grp_by_lead)
            and (not rc.filter)
            and (not rc.current)
            and (not rc.all)
            and (not rc.orphans)
        ):
            return
        if rc.date:
            now = date_parser.parse(rc.date).date()
        else:
            now = dt.date.today()
        if not rc.range:
            rc.range = 7
        since_date = now - dt.timedelta(days=int(rc.range))

        projecta, end_projecta, error_projecta = [], [], []
        grouped_projecta = {}
        if rc.grant:
            collection = [prum for prum in collection if rc.grant in prum.get("grants", [])]
        if rc.orphans:
            if rc.person or rc.lead:
                raise RuntimeError("you cannot specify lead or person with orphans")
            dead_parents = [person.get("_id") for person in self.gtx["people"] if not person.get("active")]
            collection = [
                prum
                for prum in collection
                if prum.get("lead") in dead_parents
                and prum.get("status") in PROJECTUM_ACTIVE_STATI + PROJECTUM_PAUSED_STATI
                or prum.get("lead").lower() == "tbd"
            ]
        if rc.lead:
            if rc.person or rc.orphans:
                raise RuntimeError("please specify either lead or person, not both")
            collection = [prum for prum in collection if prum.get("lead") == rc.lead]
        if rc.person:
            if rc.orphans:
                raise RuntimeError("please specify either lead or person, not both")
            if isinstance(rc.person, str):
                rc.person = [rc.person]
            collection = [
                prum
                for prum in collection
                if prum.get("lead") in rc.person
                or bool(set(prum.get("group_members", [])).intersection(set(rc.person)))
            ]
        if rc.current:
            collection = [prum for prum in collection if prum.get("status") in PROJECTUM_ACTIVE_STATI]
        elif not rc.all:
            collection = [prum for prum in collection if prum.get("status") not in PROJECTUM_CANCELLED_STATI]
        for projectum in collection:
            if rc.ended:
                if projectum.get("status") not in PROJECTUM_ACTIVE_STATI:
                    if projectum.get("status") in INACTIVE_STATI:
                        continue
                    elif projectum.get("status") not in PROJECTUM_FINISHED_STATI:
                        error_projecta.append(projectum)
                    elif not isinstance(projectum.get("end_date"), dt.date):
                        try:
                            dt.datetime(*[int(num) for num in projectum.get("end_date").split("-")])
                        except Exception:
                            error_projecta.append(projectum)
                    else:
                        end_date = projectum.get("end_date")
                        if isinstance(end_date, str):
                            end_date = date_parser.parse(end_date).date()
                        if since_date <= end_date <= now:
                            end_projecta.append(projectum)
                if end_projecta != []:
                    projecta = end_projecta
                continue
            projecta.append(projectum)

        if rc.verbose:
            for p in projecta:
                grants = None
                if p.get("grants"):
                    if isinstance(p.get("grants"), list):
                        grants = " ,".join(p.get("grants"))
                    else:
                        grants = p.get("grants")
                if p.get("status") == "finished":
                    ended = f", end_date: {p.get('end_date')}"
                else:
                    ended = ""
                print(p.get("_id"))
                print(
                    f"    status: {p.get('status')}, begin_date: {p.get('begin_date')}, "
                    f"due_date: {p.get('deliverable', {}).get('due_date')}{ended}, grant: {grants}"
                )
                print(f"    description: {p.get('description')}")
                print("    team:")
                print(f"        lead: {p.get('lead')}")
                grp_members = None
                if p.get("group_members"):
                    grp_members = ", ".join(p.get("group_members"))
                collaborators = None
                if p.get("collaborators"):
                    collaborators = ", ".join(p.get("collaborators"))
                print(f"        group_members: {grp_members}")
                print(f"        collaborators: {collaborators}")
            return

        if rc.grp_by_lead:
            for p in projecta:
                output = f"{p.get('_id')} ({p.get('status')})"
                if p.get("lead") not in grouped_projecta:
                    grouped_projecta[p.get("lead")] = [output]
                else:
                    grouped_projecta[p.get("lead")].append(output)
            for key, values in grouped_projecta.items():
                print(f"{key}:")
                for v in values:
                    print(f"    {v}")
            return

        projecta.sort(key=lambda prum: prum.get("_id"))
        if rc.keys:
            results = collection_str(projecta, rc.keys)
            print(results, end="")
            return

        if end_projecta != []:
            if now == dt.date.today() and rc.range == 7:
                print("\nProjecta finished this past week! o(*^v^*)o")
            else:
                print(f"\nProjecta finished within the {rc.range} days leading up to {now}")
        elif end_projecta == [] and rc.ended:
            if now == dt.date.today() and rc.range == 7:
                print("\nNo projecta finished this week")
            else:
                print(f"\nNo projecta finished within the {rc.range} days leading up to {now}")

        for i in projecta:
            output = f"{i.get('_id')} ({i.get('status')})"
            print(output)

        if error_projecta:
            print(
                "\nWARNING: These projecta have an issue with the end date and/or status, "
                "please run f_prum to set status to finished and add an end date"
            )
            for i in error_projecta:
                print(i.get("_id"))



"""Generic builder."""

from regolith.builders.activitylogbuilder import ActivitylogBuilder
from regolith.builders.beamplanbuilder import BeamPlanBuilder
from regolith.builders.coabuilder import RecentCollaboratorsBuilder
from regolith.builders.cpbuilder import CPBuilder
from regolith.builders.cvbuilder import CVBuilder
from regolith.builders.figurebuilder import FigureBuilder
from regolith.builders.formalletterbuilder import FormalLetterBuilder
from regolith.builders.gradebuilder import GradeReportBuilder
from regolith.builders.grantreportbuilder import GrantReportBuilder
from regolith.builders.htmlbuilder import HtmlBuilder
from regolith.builders.internalhtmlbuilder import InternalHtmlBuilder
from regolith.builders.manuscriptreviewbuilder import ManRevBuilder
from regolith.builders.postdocadbuilder import PostdocadBuilder
from regolith.builders.preslistbuilder import PresListBuilder
from regolith.builders.proposalreviewbuilder import PropRevBuilder
from regolith.builders.publistbuilder import PubListBuilder
from regolith.builders.readinglistsbuilder import ReadingListsBuilder
from regolith.builders.reimbursementbuilder import ReimbursementBuilder
from regolith.builders.resumebuilder import ResumeBuilder

BUILDERS = {
    "annual-activity": ActivitylogBuilder,
    "beamplan": BeamPlanBuilder,
    "current-pending": CPBuilder,
    "cv": CVBuilder,
    "figure": FigureBuilder,
    "formalletter": FormalLetterBuilder,
    "grade": GradeReportBuilder,
    "grades": GradeReportBuilder,
    "grant-report": GrantReportBuilder,
    "html": HtmlBuilder,
    "internalhtml": InternalHtmlBuilder,
    "postdocad": PostdocadBuilder,
    "preslist": PresListBuilder,
    "publist": PubListBuilder,
    "reading-lists": ReadingListsBuilder,
    "reimb": ReimbursementBuilder,
    "recent-collabs": RecentCollaboratorsBuilder,
    "resume": ResumeBuilder,
    "review-man": ManRevBuilder,
    "review-prop": PropRevBuilder,
}


def builder(btype, rc):
    """Returns builder of the appropriate type."""
    return BUILDERS[btype](rc)



from regolith.chained_db import ChainDB


def test_dddi():
    a = {"a": {"a": {"a": 1}}}
    z = ChainDB(a)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["a"], ChainDB)
    assert isinstance(z["a"]["a"]["a"], int)
    assert z["a"]["a"]["a"] + 1 == 2


def test_second_mapping():
    m1 = {"a": {"m": {"x": 0}}}
    m2 = {"a": {"m": {"y": 1}}}
    z = ChainDB(m1)
    z.maps.append(m2)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"].maps, list)
    assert z["a"]["m"]["y"] == 1


def test_double_mapping():
    m1 = {"a": {"m": {"y": 0}}}
    m2 = {"a": {"m": {"y": 1}}}
    z = ChainDB(m1)
    z.maps.append(m2)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"].maps, list)
    assert isinstance(z["a"]["m"]["y"], int)
    assert z["a"]["m"]["y"] == 1


def test_list_mapping():
    m1 = {"a": {"m": "x"}}
    m2 = {"a": {"m": "y"}}
    z = ChainDB(m1)
    z.maps.append(m2)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], str)
    assert z["a"]["m"] == "y"


def test_mixed_mapping():
    m1 = {"a": {"m": {"y": 1}}}
    m2 = {"a": {"m": 1}}
    z = ChainDB(m1)
    z.maps.append(m2)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], int)
    assert z["a"]["m"] == 1


def test_exactness():
    d = {"y": 1}
    m1 = {"a": {"m": d}}
    z = ChainDB(m1)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], ChainDB)
    assert isinstance(z["a"]["m"].maps[0], dict)
    assert d is z["a"]["m"].maps[0]


def test_exactness_setting():
    d = {"y": 1}
    m1 = {"a": {"m": d}}
    z = ChainDB(m1)
    e = {"z": 2}
    z["a"]["m"] = e
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], ChainDB)
    assert isinstance(z["a"]["m"].maps[0], dict)
    assert e is z["a"]["m"].maps[0]


def test_exactness_setting_multi():
    d = "a"
    e = "b"
    m1 = {"a": {"m": d}}
    m2 = {"a": {"m": e}}
    z = ChainDB(m1)
    z.maps.append(m2)
    g = ("c",)
    z["a"]["m"] = g
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], tuple)
    assert isinstance(z["a"].maps[0], dict)
    assert g is z["a"].maps[1]["m"]
    # We sent this to the first map not the last
    assert z["a"]["m"] is g


def test_exactness_setting_multi2():
    d = [1, 2]
    e = [3, 4]
    ee = [5, 6]
    m1 = {"a": {"m": d}}
    m2 = {"a": {"m": e, "mm": ee}}
    z = ChainDB(m1)
    z.maps.append(m2)
    g = [-1, -2]
    z["a"]["mm"] = g
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], list)
    assert isinstance(z["a"].maps[0], dict)
    assert g is z["a"].maps[1]["mm"]
    assert z["a"]["mm"] == g
    assert z["a"]["m"] == [1, 2, 3, 4]


def test_exactness_setting_multi_novel():
    d = [1, 2]
    e = [3, 4]
    m1 = {"a": {"m": d}}
    m2 = {"a": {"m": e}}
    z = ChainDB(m1)
    z.maps.append(m2)
    g = [-1, -2]
    z["a"]["mm"] = g
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["m"], list)
    assert isinstance(z["a"].maps[0], dict)
    assert g is z["a"].maps[0]["mm"]
    assert g == z["a"]["mm"]


def test_dicts_in_lists():
    c = [{"m": 1}, {"n": 2}]
    d = [{"o": 3}, {"p": 4}]
    t = c + d
    m1 = {"a": {"b": c}}
    m2 = {"a": {"b": d}}
    z = ChainDB(m1)
    z.maps.append(m2)
    assert isinstance(z["a"], ChainDB)
    assert isinstance(z["a"]["b"], list)
    assert z["a"]["b"] == t
    assert c[0] is z["a"]["b"][0]
    assert c[1] is z["a"]["b"][1]
    assert d[0] is z["a"]["b"][2]
    assert d[1] is z["a"]["b"][3]


def test_dicts_in_lists_mutation():
    c = [{"m": 1}, {"n": 2}]
    d = [{"o": 3}, {"p": 4}]
    m1 = {"a": {"b": c}}
    m2 = {"a": {"b": d}}
    z = ChainDB(m1)
    z.maps.append(m2)
    append_list = z["a"]["b"]
    append_list.append({"hi": "world"})
    assert isinstance(append_list, list)
    assert z["a"]["b"] != append_list

    extend_list = z["a"]["b"]
    extend_list.extend([{"hi": "world"}, {"spam": "eggs"}])
    assert z["a"]["b"] != extend_list



"""Builder for websites."""

import datetime as dt
import os
import shutil

from regolith.builders.basebuilder import BuilderBase
from regolith.dates import get_dates
from regolith.fsclient import _id_key
from regolith.sorters import ene_date_key, position_key
from regolith.tools import (
    all_docs_from_collection,
    dereference_institution,
    document_by_value,
    filter_projects,
    filter_publications,
    fuzzy_retrieval,
    get_formatted_crossref_reference,
    make_bibtex_file,
)

PROJ_URL_BASE = "https://gitlab.thebillingegroup.com/talks/"


class InternalHtmlBuilder(BuilderBase):
    """Build HTML files for website"""

    btype = "internalhtml"
    needed_colls = ["people", "meetings"]

    def __init__(self, rc):
        super().__init__(rc)
        # TODO: get this from the RC
        self.cmds = [
            "root_index",
            "meetings",
            "nojekyll",
            "cname",
        ]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["jobs"] = list(all_docs_from_collection(rc.client, "jobs"))
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["meetings"] = list(all_docs_from_collection(rc.client, "meetings"))
        gtx["group"] = document_by_value(all_docs_from_collection(rc.client, "groups"), "name", rc.groupname)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["institutions"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)

    def finish(self):
        """Move files over to their destination and remove them from the
        source"""
        # static
        stsrc = os.path.join(getattr(self.rc, "static_source", "templates"), "static")
        stdst = os.path.join(self.bldir, "static")
        if os.path.isdir(stdst):
            shutil.rmtree(stdst)
        shutil.copytree(stsrc, stdst)

    def root_index(self):
        """Render root index"""
        self.render("introot_index.html", "intindex.html", title="Home")

    def meetings(self):
        """Render projects"""
        rc = self.rc
        mtgsi = all_docs_from_collection(rc.client, "meetings")

        pp_mtgs, f_mtgs, jclub_cumulative = [], [], []
        for mtg in mtgsi:
            if not mtg.get("lead"):
                print("{} missing a meeting lead".format(mtg["_id"]))
            if not mtg.get("scribe"):
                print("{} missing a meeting scribe".format(mtg["_id"]))
            lead = fuzzy_retrieval(
                all_docs_from_collection(rc.client, "people"), ["_id", "name", "aka"], mtg.get("lead")
            )
            if not lead:
                print("{} lead {} not found in people".format(mtg["_id"], mtg.get("lead")))
            mtg["lead"] = lead["name"]
            scribe = fuzzy_retrieval(
                all_docs_from_collection(rc.client, "people"), ["_id", "name", "aka"], mtg.get("scribe")
            )
            if not scribe:
                print("{} scribe {} not found in people".format(mtg["_id"], mtg.get("scribe")))
            mtg["scribe"] = scribe["name"]
            if mtg.get("journal_club"):
                prsn_id = mtg["journal_club"].get("presenter", "None")
                prsn = fuzzy_retrieval(
                    all_docs_from_collection(rc.client, "people"), ["_id", "name", "aka"], prsn_id
                )
                if not prsn:
                    if prsn_id.lower() not in ["tbd", "hold", "na"]:
                        print(
                            "WARNING: {} presenter {} not found in people".format(
                                mtg["_id"], mtg["presentation"].get("presenter")
                            )
                        )
                    prsn = {"name": prsn_id}
                mtg["journal_club"]["presenter"] = prsn["name"]
                mtg_jc_doi = mtg["journal_club"].get("doi", "tbd")
                mtg_jc_doi_casefold = mtg_jc_doi.casefold()
                if mtg_jc_doi_casefold == "na":
                    mtg["journal_club"]["doi"] = "N/A"
                elif mtg_jc_doi_casefold != "tbd":
                    if not mtg_jc_doi_casefold.startswith("arxiv"):
                        ref, _ = get_formatted_crossref_reference(mtg["journal_club"].get("doi"))
                        mtg["journal_club"]["doi"] = ref
                    else:
                        ref = mtg_jc_doi
                    jclub_cumulative.append((ref, get_dates(mtg).get("date", "no date")))

            if mtg.get("presentation"):
                prsn_id = mtg["presentation"].get("presenter", "None")
                prsn = fuzzy_retrieval(
                    all_docs_from_collection(rc.client, "people"), ["_id", "name", "aka"], prsn_id
                )
                if not prsn:
                    if prsn_id.lower() not in ["tbd", "hold", "na"]:
                        print(
                            "WARNING: {} presenter {} not found in people".format(
                                mtg["_id"], mtg["presentation"].get("presenter")
                            )
                        )
                    prsn = {"name": prsn_id}
                mtg["presentation"]["presenter"] = prsn["name"]
                mtg["presentation"]["link"] = mtg["presentation"].get("link", "tbd")
            mtg["date"] = dt.date(mtg.get("year"), mtg.get("month"), mtg.get("day"))
            mtg["datestr"] = mtg["date"].strftime("%m/%d/%Y")
            today = dt.date.today()
            if mtg["date"] >= today:
                f_mtgs.append(mtg)
            else:
                pp_mtgs.append(mtg)

        jclub_cumulative = sorted(jclub_cumulative, key=lambda x: x[1], reverse=True)
        pp_mtgs = sorted(pp_mtgs, key=lambda x: x.get("date"), reverse=True)
        f_mtgs = sorted(f_mtgs, key=lambda x: x.get("date"))
        self.render(
            "grpmeetings.html",
            "grpmeetings.html",
            title="Group Meetings",
            ppmeetings=pp_mtgs,
            fmeetings=f_mtgs,
            jclublist=jclub_cumulative,
        )

    def nojekyll(self):
        """Touches a nojekyll file in the build dir"""
        with open(os.path.join(self.bldir, ".nojekyll"), "a+"):
            pass

    def cname(self):
        """Add CNAME"""
        rc = self.rc
        if not hasattr(rc, "cname"):
            return
        with open(os.path.join(self.bldir, "CNAME"), "w", encoding="utf-8") as f:
            f.write(rc.cname)

    def people(self):
        """Render people, former members, and each person"""
        rc = self.rc
        peeps_dir = os.path.join(self.bldir, "people")
        former_peeps_dir = os.path.join(self.bldir, "former")
        os.makedirs(peeps_dir, exist_ok=True)
        os.makedirs(former_peeps_dir, exist_ok=True)
        peeps = self.gtx["people"]
        for p in peeps:
            names = frozenset(p.get("aka", []) + [p["name"]])
            pubs = filter_publications(
                all_docs_from_collection(rc.client, "citations"),
                names,
                reverse=True,
                bold=False,
            )

            bibfile = make_bibtex_file(pubs, pid=p["_id"], person_dir=peeps_dir)
            ene = p.get("employment", []) + p.get("education", [])
            ene.sort(key=ene_date_key, reverse=True)
            for e in ene:
                dereference_institution(e, self.gtx["institutions"])
            projs = filter_projects(all_docs_from_collection(rc.client, "projects"), names)
            self.render(
                "person.html",
                os.path.join("people", p["_id"] + ".html"),
                p=p,
                title=p.get("name", ""),
                pubs=pubs,
                names=names,
                bibfile=bibfile,
                education_and_employment=ene,
                projects=projs,
            )
        self.render("people.html", os.path.join("people", "index.html"), title="People")

        self.render(
            "former.html",
            os.path.join("former", "index.html"),
            title="Former Members",
        )

    def projects(self):
        """Render projects"""
        rc = self.rc
        projs = all_docs_from_collection(rc.client, "projects")
        self.render("projects.html", "projects.html", title="Projects", projects=projs)

    def blog(self):
        """Render the blog and rss"""
        rc = self.rc
        blog_dir = os.path.join(self.bldir, "blog")
        os.makedirs(blog_dir, exist_ok=True)
        posts = list(all_docs_from_collection(rc.client, "blog"))
        posts.sort(key=ene_date_key, reverse=True)
        for post in posts:
            self.render(
                "blog_post.html",
                os.path.join("blog", post["_id"] + ".html"),
                post=post,
                title=post["title"],
            )
        self.render(
            "blog_index.html",
            os.path.join("blog", "index.html"),
            title="Blog",
            posts=posts,
        )
        self.render("rss.xml", os.path.join("blog", "rss.xml"), items=posts)

    def jobs(self):
        """Render the jobs and each job"""
        jobs_dir = os.path.join(self.bldir, "jobs")
        os.makedirs(jobs_dir, exist_ok=True)
        for job in self.gtx["jobs"]:
            self.render(
                "job.html",
                os.path.join("jobs", job["_id"] + ".html"),
                job=job,
                title="{0} ({1})".format(job["title"], job["_id"]),
            )
        self.render("jobs.html", os.path.join("jobs", "index.html"), title="Jobs")

    def abstracts(self):
        """Render each abstract"""
        abs_dir = os.path.join(self.bldir, "abstracts")
        os.makedirs(abs_dir, exist_ok=True)
        for ab in self.gtx["abstracts"]:
            self.render(
                "abstract.html",
                os.path.join("abstracts", ab["_id"] + ".html"),
                abstract=ab,
                title="{0} {1} - {2}".format(ab["firstname"], ab["lastname"], ab["title"]),
            )



# stup import hooks
import xonsh.imphooks
from xonsh.built_ins import XSH
from xonsh.execer import Execer

XSH.load(execer=Execer())
execer = XSH.execer
xonsh.imphooks.install_import_hooks(execer=execer)

__version__ = "0.8.2"

del xonsh



"""Emails people via SMTP"""

import os
import smtplib
import tempfile
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

try:
    from docutils.core import publish_string
except ImportError:
    publish_string

from regolith.builders.gradebuilder import GradeReportBuilder
from regolith.tools import all_docs_from_collection


def attach_txt(filename):
    with open(filename, "r", encoding="utf-8") as f:
        txt = f.read()
    msg = MIMEText(txt, _subtype="text")
    return msg


def attach_pdf(filename):
    with open(filename, "rb", encoding="utf-8") as f:
        pdf = f.read()
    msg = MIMEApplication(pdf, _subtype="pdf")
    return msg


ATTACHERS = {
    ".txt": attach_txt,
    ".rst": attach_txt,
    ".pdf": attach_pdf,
    ".ipynb": attach_txt,
}


def make_message(rc, to, subject="", body="", attachments=()):
    """Creates an email following the approriate format. The body kwarg
    may be a string of restructured text.  Attachements is a list of filenames
    to attach.
    """
    msg = MIMEMultipart("alternative")
    plain = MIMEText(body, "plain")
    msg.attach(plain)
    if publish_string is not None:
        html = publish_string(
            body,
            writer_name="html",
            settings_overrides={"output_encoding": "unicode"},
        )
        html = MIMEText(html, "html")
        msg.attach(html)
    if attachments:
        text = msg
        msg = MIMEMultipart("mixed")
        msg.attach(text)
        for attachment in attachments:
            _, ext = os.path.splitext(attachment)
            att = ATTACHERS[ext](attachment)
            att.add_header(
                "content-disposition",
                "attachment",
                filename=os.path.basename(attachment),
            )
            msg.attach(att)
    msg["Subject"] = subject
    msg["From"] = rc.email["from"]
    msg["To"] = to
    return (to, msg.as_string())


def test_email(rc):
    """Sends a test email from regolith."""
    if rc.to is None:
        raise ValueError("--to must be given to send a test email.")
    with tempfile.NamedTemporaryFile(suffix=".rst", delete=False) as f:
        f.write(b"This is *only* a test attachment.\n")
        f.flush()
        message = make_message(
            rc,
            rc.to,
            subject="Regolith test email",
            body="This is *only* a test. Do not be alarmed.",
            attachments=[f.name],
        )
    return [message]


def grade_email(rc):
    """Sends grade report emails to students."""
    gradedir = os.path.join(rc.builddir, GradeReportBuilder.btype)
    addresses = {x["_id"]: x["email"] for x in list(all_docs_from_collection(rc.client, "students"))}
    messages = []
    for course in all_docs_from_collection(rc.client, "courses"):
        if not course.get("active", True):
            continue
        course_id = course["_id"]
        if course_id not in rc.course_ids:
            continue
        for student_id in course["students"]:
            base = GradeReportBuilder.basename(student_id, course_id) + ".pdf"
            fname = os.path.join(gradedir, base)
            if not os.path.isfile(fname):
                raise RuntimeError(
                    fname + " does not exist, please run " '"regolith build grade" prior to emailing ' "grades."
                )
            message = make_message(
                rc,
                addresses[student_id],
                subject="Current grades for " + course_id,
                body="Please see the attached PDF and " "please report any errors.",
                attachments=[fname],
            )
            messages.append(message)
    return messages


def class_email(rc):
    """Sends an email to all students in the active classes."""
    addresses = {x["_id"]: x["email"] for x in list(all_docs_from_collection(rc.client, "students"))}
    messages = []
    for course in all_docs_from_collection(rc.client, "courses"):
        if not course.get("active", True):
            continue
        course_id = course["_id"]
        if course_id not in rc.course_ids:
            continue
        subject = "[{0}] {1}".format(course_id, rc.subject)
        for student_id in course["students"]:
            message = make_message(
                rc,
                addresses[student_id],
                subject=subject,
                body=rc.body,
                attachments=rc.attachments,
            )
            messages.append(message)
    return messages


def list_email(rc):
    """List class emails"""
    course = rc.client.find_one(rc.db, "courses", {"_id": rc.course_ids})
    student_ids = set(course["students"])
    students = rc.client[rc.db]["students"]
    emails = [s["email"] for s in students.values() if s["_id"] in student_ids]
    print(",".join(sorted(emails)))


EMAIL_CONSTRUCTORS = {
    "test": test_email,
    "grade": grade_email,
    "grades": grade_email,
    "class": class_email,
    "list": list_email,
}


def emailer(rc):
    """Constructs and sends out emails"""
    constructor = EMAIL_CONSTRUCTORS[rc.email_target]
    emails = constructor(rc)
    if emails is None:
        return
    conf = rc.email
    with smtplib.SMTP(conf["url"], port=conf["port"]) as smtp:
        smtp.set_debuglevel(conf["verbosity"])
        if conf["tls"]:
            smtp.starttls()
            smtp.ehlo()
        smtp.login(conf["user"], conf["password"])
        for to, message in emails:
            print("sending email to " + to + "...")
            smtp.sendmail(conf["from"], to, message)



"""Helper for marking a task as finished in todos collection."""

import datetime as dt
import math

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import (
    all_docs_from_collection,
    document_by_value,
    get_pi_id,
    key_value_pair_filter,
    print_task,
)

TARGET_COLL = "todos"
ALLOWED_IMPORTANCE = [0, 1, 2]


def subparser(subpi):
    date_kwargs = {}
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"
        int_kwargs["widget"] = "IntegerField"

    subpi.add_argument(
        "-i",
        "--index",
        help="Enter the index of a certain task in the enumerated list to mark as finished.",
        type=int,
    )
    subpi.add_argument("--end-date", help="Add the end date of the task. Default is today.", **date_kwargs)
    subpi.add_argument(
        "-t",
        "--assigned-to",
        help="Filter tasks that are assigned to this user id. Default id is saved in user.json. ",
    )
    subpi.add_argument(
        "-b",
        "--assigned-by",
        nargs="?",
        const="default_id",
        help="Filter tasks that are assigned to other members by this user id. Default id is saved in user.json. ",
    )
    subpi.add_argument(
        "-f",
        "--filter",
        nargs="+",
        help="Search this collection by giving key element pairs. "
        "'-f description paper' will return tasks with description containing 'paper' ",
    )
    subpi.add_argument(
        "--date",
        help="Enter a date such that the helper can calculate how many days are left "
        "from that date to the due-date. Default is today.",
        **date_kwargs,
    )
    return subpi


class TodoFinisherHelper(DbHelperBase):
    """Helper for marking a task as finished in todos collection."""

    # btype must be the same as helper target in helper.py
    btype = "f_todo"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)

        rc.coll = f"{TARGET_COLL}"
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        if rc.index:
            if rc.index >= 9900:
                print(
                    "WARNING: indices >= 9900 are used for milestones which "
                    "should be finished using u_milestone and not f_todo"
                )
                return
        if not rc.assigned_to:
            try:
                rc.assigned_to = rc.default_user_id
            except AttributeError:
                print(
                    "Please set default_user_id in '~/.config/regolith/user.json', "
                    "or you need to enter your group id in the command line"
                )
                return
        person = document_by_value(all_docs_from_collection(rc.client, "todos"), "_id", rc.assigned_to)
        filterid = {"_id": rc.assigned_to}
        if not person:
            raise TypeError(f"Id {rc.assigned_to} can't be found in todos collection")
        todolist = person.get("todos", [])
        if len(todolist) == 0:
            print(f"{rc.assigned_to} doesn't have todos in todos collection.")
            return
        now = dt.date.today()
        if not rc.index:
            if not rc.date:
                today = now
            else:
                today = date_parser.parse(rc.date).date()
            if rc.filter:
                todolist = key_value_pair_filter(todolist, rc.filter)
            for todo in todolist:
                if isinstance(todo["due_date"], str):
                    todo["due_date"] = date_parser.parse(todo["due_date"]).date()
                todo["days_to_due"] = (todo.get("due_date") - today).days
                todo["order"] = 1 / (1 + math.exp(abs(todo["days_to_due"] - 0.5)))
            todolist = sorted(
                todolist, key=lambda k: (k["status"], k["importance"], k["order"], -k.get("duration", 10000))
            )
            print(
                "If the indices are far from being in numerical order, "
                "please renumber them by running regolith helper u_todo -r"
            )
            print("Please choose from one of the following to update:")
            print("(index) action (days to due date|importance|expected duration (mins)|tags|assigned by)")
            print("-" * 80)
            print_task(todolist, stati=["started"])
        else:
            match_todo = [i for i in todolist if i.get("running_index") == rc.index]
            if len(match_todo) == 0:
                raise RuntimeError("Please enter a valid index.")
            else:
                todo = match_todo[0]
                todo["status"] = "finished"
                if not rc.end_date:
                    end_date = now
                else:
                    end_date = date_parser.parse(rc.end_date).date()
                todo["end_date"] = end_date
                for i in range(0, len(rc.databases)):
                    db_name = rc.databases[i]["name"]
                    person_update = rc.client.find_one(db_name, rc.coll, filterid)
                    if person_update:
                        todolist_update = person_update.get("todos", [])
                    else:
                        continue
                    if len(todolist_update) != 0:
                        for i, todo_u in enumerate(todolist_update):
                            if rc.index == todo_u.get("running_index"):
                                todolist_update[i] = todo
                                rc.client.update_one(
                                    db_name,
                                    rc.coll,
                                    {"_id": rc.assigned_to},
                                    {"todos": todolist_update},
                                    upsert=True,
                                )
                                print(
                                    f"The task \"({todo_u['running_index']}) {todo_u['description'].strip()}\" "
                                    f"in {db_name} for {rc.assigned_to} has been marked as finished."
                                )
                                return
        return
